/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************
NAME            mppLnx_vhbalib.c
SUMMARY         %description%
VERSION         %version: 15 %
UPDATE DATE     %date_modified: Fri Feb 06 13:29:52 2004 %
PROGRAMMER      %created_by:    sdachepa %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:
   This function contains the functions in the Linux MPP virtual HBA that manages
   the virtual HBA's internal data model. The functions include creation/deletion
   of list entries of different entry structs and manage the relations of virtual 
   SCSI commands and physical SCSI commands.

INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:
  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

#define __SRCmppLnx_vhbalib_c

/***  INCLUDES  ***/  
/***  Linux kernel INCLUDES  ***/  
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <asm/io.h>

#include <linux/blk.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"

#include <linux/stat.h>

/***  MPP commomn API INCLUDES  ***/  
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "MPP_Sysdep.h"
#include "mppLnx_version.h"
/***  Linux MPP Virtual HBA INCLUDES  ***/  
#include "mppLnx_vhba.h"
#include "mppLnx_shared.h"


/***  CONSTANT DEFINITIONS  ***/


/***  MACRO DEFINITIONS  ***/


/***  TYPE DEFINITIONS  ***/


/***  LOCALS  ***/
mppLnx_VHBAContext_t mppLnxVHBAContext;
/*
*declare a spinlock for protecting the context of the mppLnxVHBAContext
*/
static spinlock_t mppLnxVHBAContext_Lock = SPIN_LOCK_UNLOCKED;

/***  PROCEDURES  ***/



/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_initVHBAContext
* SUMMARY:  Initialize the internal data model of the Linux MPP virtual HBA 
         driver module.
* SCOPE:   local
*
*
* DESCRIPTION:
      This function is invoked by the mppLnx_detect() function, which is the 
      first function in the Linux MPP virtual HBA driver being invoked by the
      Linux SCSI middle level. This function initializes the internal data model, 
      walks through the MPP common data model and creates entry lists of the 
      virtual HBA's data model and /proc entries for all virtual and physical devices
* RESTRICTIONS:
* 
* RETURNS:0 indicates the proc leaf node is successfully 
*           created or 1 otherwise
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_initVHBAContext()
{
   int                           vTarget =0;
   int                           vLun = 0;
   int                           cntl = 0;
   int                           path = 0;
   int rtn  = 0;
   RdacDeviceInformation_t       *rdacInfo;
   RdacLunInformation_t          *rdaclun;
   LunPathObjects_t              * lunPathObject;
   Scsi_Device                   *sdp;

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
      "Entering mppLnx_initVHBAContext()\n"));
   /*
   * assign the lock to the data struct
   */
   mppLnxVHBAContext.vhbaContextLock = &mppLnxVHBAContext_Lock;
   /*
   * insert physical devices to the internal data struct - mppLnxVHBAContext
   */
   for(vTarget = 0; vTarget < mppLnx_Config_Parameters.mpp_MaxArrayModules; vTarget++)
   {
      if( (rdacInfo = mpp_ModuleArray[vTarget].RdacInfo) == NULL)
      {
         MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
            "no storage array at virtual target %d\n",vTarget));
         continue;
      }
      
      for(vLun = 0; vLun < mppLnx_Config_Parameters.mpp_MaxLunsPerArray; vLun++)
      {
         rdaclun = &(rdacInfo->RdacLuns[vLun]);
         if(!rdaclun->Present)
         {
            MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
               "No device at the addresses virtual target %d lun %d\n",
               vTarget, vLun));
            continue;
         }

         /*
         * we must create virtual device proc entry here. This is because the proc
         * API create_proc_read_entry() allocates memory with GFP_KERNEL flag. Therefore
         * we cannot create it in interrupt handling.
         * The problem is that at this time, the virtual Scsi_Device has not being
         * created yet. We must take care of the virtual Scsi_Device pointer later.
         */

         if(!rdaclun->UTMLun)
         {
            mppLnx_insert_VdeviceEntry(vLun, rdacInfo);
         }else
         {
            MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
               "Don't add UTM virtual LUN to virtual device entry %d lun %d\n",
               vTarget, vLun));
         }

         for(cntl = 0; cntl < 2; cntl++)
         {
            lunPathObject = rdaclun->LunControllers[cntl];
            for(path = 0; path < mppLnx_Config_Parameters.mpp_MaxPathsPerController; path++)
            {
               sdp = lunPathObject->LunPaths[path].LunPathDevice;
               if(sdp != NULL)
               {
                  MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
                     "found a physical device at h%dc%dt%dl%d\n",
                     sdp->host->host_no, sdp->channel, sdp->id, sdp->lun));
                  MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
                     "From array %s controller %d path %d\n",
                     (char *)rdacInfo->ModuleName, cntl,path));
                  mppLnx_insert_PdeviceEntry(sdp, rdacInfo, cntl, path);
               }
            }
         }
      }
   }
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_initVHBAContext()\n"));
   return rtn;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_cleanVHBAContext
* SUMMARY:  Cleans up the Linux MPP virtual HBA driver internal data model and 
      release system resources.
* SCOPE:   public
*
*
* DESCRIPTION:
   This function is invoked by the mppLnx_release() function, which is the last
   function of the Linux MPP virtual HBA driver being invoked by the Linux SCSI
   middle level at the time of the virtual HBA driver module unloading time.

   This function cleans-up the Linux MPP virtual HBA driver's internal data model 
   and release system resources. The system resources include the memory allocated
   for the internal model, and the /proc entries of virtual devices and physical
   devices.

* RESTRICTIONS:
* 
* RETURNS:SUCCESS indicates the proc leaf node is successfully 
*           created or FAILED otherwise
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_cleanVHBAContext()
{
   int                        lun = 0;
   mppLnx_VDeviceEntry_t      *vde;
   struct list_head           *listPt;
   struct list_head           *phlistPt;
   struct list_head           *ptlistPt;

   mppLnx_PDeviceEntry_t      *pde;
   mppLnx_PhostEntry_t        *phe;
   mppLnx_PtargetEntry_t      *pte;
   unsigned long              iflag;
   
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_cleanVHBAContext()\n"));
   /**
   * remove/free virtual device list 
   */
   if(NULL == mppLnxVHBAContext.virtualDeviceList )
   {
      /* no virtual device */
      MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
         "No virtual device in the virtual device list\n"));
   }
   else
   {
      while(1)
      {
         spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
         if( (listPt = mppLnxVHBAContext.virtualDeviceList->list.next)
             != &(mppLnxVHBAContext.virtualDeviceList->list))
         {
            vde = list_entry(listPt, mppLnx_VDeviceEntry_t, list);
            spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
               MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
               "Deleting the virtual device entry %d:%d\n",
               vde->rdacInfo->VirtualTargetId,
               vde->lun));
            mppLnx_delete_VdeviceEntry(vde);
         }
         else
         {
            /* finally, release the empty list header itself */
            kfree(mppLnxVHBAContext.virtualDeviceList);
            mppLnxVHBAContext.virtualDeviceList = NULL;
            spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
            break;
         }
      }
   }
   /*
   * remove/free physical device list
   */
   if(NULL == mppLnxVHBAContext.phostList )
   {
      /* no physical device in the internal data model */
      MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
         "No physical device in the internal data model\n"));
   }else
   {
      while(1)
      {
         /*
         * all physical Scsi_Host entry
         */
         spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
         if((phlistPt = mppLnxVHBAContext.phostList->list.next)
            != &(mppLnxVHBAContext.phostList->list))
         {
            phe = list_entry(phlistPt, mppLnx_PhostEntry_t, list);
            spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
            MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
               "De-initializing on host %s hostNo %d\n",
               phe->host->hostt->proc_name, phe->hostNo));
            /*
            * physical target entry
            */
            spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
            if((ptlistPt = phe->targetList->list.next)
               != &(phe->targetList->list))
            {
               pte = list_entry(ptlistPt, mppLnx_PtargetEntry_t, list);
               spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);

               MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
                  "De-initializing target %d\n", pte->targetId));
               /*
               * Lun entry of each target
               */
               for(lun = 0; lun < mppLnx_Config_Parameters.mpp_MaxLunsPerArray; lun++)
               {
                  pde = pte->pdevices[lun];
                  if(pde != NULL)
                  {
                     MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
                        "deleting mppLnx_PDeviceEntry h%dc%dt%dl%d\n",
                        pde->pdevice->host->host_no,
                        pde->pdevice->channel,
                        pde->pdevice->id,
                        pde->pdevice->lun));
                     mppLnx_delete_PdeviceEntry(pde->pdevice);
                  }
                  else
                  {
                     MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
                        "No mppLnx_PDeviceEntry at targetId %d lun %d\n",
                        pte->targetId, lun));
                  }
               } /* for lun loop */
               spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
               list_del(&(pte->list)); /* delete the target from the list */
               spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
               kfree(pte->pdevices); /* free its pdevice ( array of pointers */
               kfree(pte); /* free the mppLnx_PtargetEntry_t itself */
            }/* for target if */
            else
            {
            
               /* free the mppLnx_PhostEntry_t's targetList */
               kfree(phe->targetList);
               /* delete the mppLnx_PhostEntry_t object from its list */
               list_del(&(phe->list)); 
               spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
               /* free the memory of the mppLnx_PhostEntry_t object */
               kfree(phe);
            }

         } /* for host loop*/ 
         else
         {
            spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
            kfree(mppLnxVHBAContext.phostList); /* free the empty list head*/
            break; /*break the while (1) */
         }
      }/* while(1) */
      
   }

   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_cleanVHBAContext()\n"));
   return 0;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_get_VdeviceEntry
* SUMMARY:  Gets a virtual device entry object by a virtual device
* SCOPE:    public
*
* DESCRIPTION:
   This function looks up the mppLnx_VHBAContext for a mppLnx_VdeviceEntry_t 
   object that represent the virtual Scsi_Device object of "vdevice".
* @vdevice A virtual Scsi_Device object whose mppLnx_VDeviceEntry_t is requested.
* RESTRICTIONS:
* 
* RETURNS: Returns a mppLnx_VDeviceEntry_t object that represents the virtual
            scsi device or null if fails.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_VDeviceEntry_t * mppLnx_get_VdeviceEntry(Scsi_Device *vdevice)
{
   mppLnx_VDeviceEntry_t   *vDEp = NULL;
   struct list_head        * listPt;
   unsigned long           iflag;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Entering mppLnx_get_VdeviceEntry()\n"));

   if(NULL == vdevice)
   {
      MPPLNX_VHBA_BUG();
   }

   if(NULL == mppLnxVHBAContext.virtualDeviceList)
   {
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "Exiting mppLnx_get_VdeviceEntry()- Null virtualDeviceList \n"));
      return NULL;

   }
   else
   {
      spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
      for(listPt = mppLnxVHBAContext.virtualDeviceList->list.next;
         listPt != &(mppLnxVHBAContext.virtualDeviceList->list);
         listPt = listPt->next)
      {
         vDEp = list_entry(listPt, mppLnx_VDeviceEntry_t, list);
         if(vDEp->rdacInfo->VirtualTargetId == vdevice->id &&
            vDEp->lun == vdevice->lun)
         {
            /*
            * If this function is called first time from a virtual command
            * completion function, the vdevuce is not set yet. The vdevice 
            * is used later, we must set it here.
            */
            if(vDEp->vdevice == NULL)
            {
               vDEp->vdevice = vdevice;
            }
            break;
         }
         else
         {
            vDEp = NULL;
         }
      }
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   }
  
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_getVdeviceEntry()\n"));
   return vDEp;
     
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_insert_VdeviceEntry
* SUMMARY:  creates and insert a mppLnx_VDeviceEntry_t object to mppLnx_VHNAContext
* SCOPE:   public
*
*
* DESCRIPTION:
   This function creates a "mppLnx_VdeviceEntry_t object" based on the input arguments
   and inserts the object to the MPP virtual HBA driver's internal data model.
   If an object mppLnx_VdeviceEntry_t, which represents the input "vdevice",  already 
   exists in the internal data model, this function doesn't insert the object 
   to the list again.

*  @lun the LUN number of a  virtual Scsi_Device object to be inserted
*  @rdacInfo a RdacDevuceInformation_t object that represents a storage array where the
*           vdevice is a virtual LUN of the array* 
RESTRICTIONS:
* 
* RETURNS: An mppLnx_VdeviceEntry_t object
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_VDeviceEntry_t * mppLnx_insert_VdeviceEntry
(
   int                        lun, 
   RdacDeviceInformation_t    * rdacInfo
)
{
   mppLnx_VDeviceEntry_t            *vde;
   struct list_head                 *listp;
   int                              size = 0;
   unsigned long                    iflag;
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Enterting mppLnx_insert_VdeviceEntry()\n"));

   if(NULL == rdacInfo)
   {
      MPPLNX_VHBA_BUG();
   }

   if(mppLnxVHBAContext.virtualDeviceList == NULL)
   {
      /*
      * create an empty entry 
      */
      size = sizeof(mppLnx_VDeviceEntry_t);
      vde = kmalloc(size, GFP_ATOMIC);
      if(NULL == vde)
      {
         /* error. Cannot get kernel memory for mppLnx_VDeviceEntry_t*/
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 420,
            "Cannot get kernel memory for mppLnx_VDeviceEntry_t\n"));
         return NULL;
      }
      memset(vde, 0 , size);
      mppLnxVHBAContext.virtualDeviceList = vde;
	   listp = &vde->list;
      INIT_LIST_HEAD(listp);
   }

   /*
   * check whether or not the entry is already in the list
   */
   vde = NULL;
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   for(listp = mppLnxVHBAContext.virtualDeviceList->list.next;
      listp != &(mppLnxVHBAContext.virtualDeviceList->list);
      listp = listp->next)
   {
      vde = list_entry(listp, mppLnx_VDeviceEntry_t, list);
      if(vde->rdacInfo->VirtualTargetId == rdacInfo->VirtualTargetId &&
         vde->lun == lun)
      {
         MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
            "Found the virtual device entry for vTarget %d vLun %d\n",
            rdacInfo->VirtualTargetId, lun));
         break;
      }else
      {
         vde = NULL;
      }
   }
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);


   /*
   * it is not in. Creates it and adds it to the list
   */
   if(NULL == vde)
   {
      /* this is a new vritual device */
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         " add a new virtual device to virtual device list vTarget %d vLun %d\n",
            rdacInfo->VirtualTargetId, lun));
      vde = mppLnx_createVDeviceEntry(lun, rdacInfo);
      spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
      list_add_tail(&(vde->list), &(mppLnxVHBAContext.virtualDeviceList->list));
      
      vde->rdacInfo = rdacInfo;
      vde->ioCount = 0;
      vde->longestTrip = 0;
      vde->shortestTrip = 0;
      vde->numCntlFailures = 0;
      vde->numPathFailures = 0;
      vde->lun = lun;
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);

      mppLnx_createVDeviceProcNode(vde);
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "mppLnx_insert_VdeviceEntry() creates a new mppLnx_VDeviceEntry_t entry\n"));
   }
   else
   {
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "mppLnx_insert_VdeviceEntry() the entry is already in the list\n"));
   }
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_insert_VdeviceEntry()\n"));
   
   return vde;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_delete_VdeviceEntry
* SUMMARY:  remove a mppLnx_VDeviceEntry_t object from mppLnx_VHNAContext,
            delete its proc node, and free its memory
* SCOPE:   public
*
* DESCRIPTION:
   This function remove a "mppLnx_VdeviceEntry_t object", based on the input
   Scsi virtual device, from the MPP virtual HBA driver's internal data
   model.
   It removes this mppLnx_VDeviceEntry_t object's /proc file and frees this 
   object's memory.
*  @vdevice a virtual Scsi_Device object to be deleted.

* RESTRICTIONS:
* 
* RETURNS: 0 if the resources are successfully released. Otherwise 1.
*
* ERRNO:
*
* NOTES:
*       This function performs a reverse operation as the function of
*       mppLnx_insert_VdeviceEntry does
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_delete_VdeviceEntry(mppLnx_VDeviceEntry_t * vde)
{
   unsigned long              iflag;

   if(NULL == vde)
   {
      MPPLNX_VHBA_BUG();
   }

   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_delete_VdeviceEntry() %s:Lun %d\n",
      (char *)vde->rdacInfo->ModuleName,vde->lun));

   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   list_del(&(vde->list));
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   /*
    * remove its proc node
    */
    mppLnx_removeVDeviceProcNode(vde);
   /*
   * free its memory
   */
   kfree(vde);

   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_delete_VdeviceEntry()\n"));   
   return 0;
}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_get_PdeviceEntry
* SUMMARY:  Gets a physical device entry object by a physical device
* SCOPE:   public
*
*
* DESCRIPTION:
   This function looks up the mppLnx_VHBAContext for a mppLnx_PdeviceEntry_t 
   object that represent the physical Scsi_Device object of "pdevice".
* @pdevice A physical Scsi_Device object whose mppLnx_PDeviceEntry_t object is
*          requested.
* RESTRICTIONS:
* 
* RETURNS:Returns a mppLnx_PDeviceEntry_t object that represents the physical
            scsi device or null if fails.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_PDeviceEntry_t * mppLnx_get_PdeviceEntry(Scsi_Device *pdevice)
{
   int                        hostNo = 0;
   int                        targetId =0;
   int                        lun = 0;
   struct                     list_head * listPrt;
   mppLnx_PhostEntry_t        * pHostEntry;
   mppLnx_PtargetEntry_t      * pTargetEntry=NULL;
   mppLnx_PDeviceEntry_t      * pDeviceEntry;
   unsigned long              iflag;
  
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Entering mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
      pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun));
   
   hostNo = pdevice->host->host_no;
   targetId = pdevice->id;
   lun = pdevice->lun;
 
   if(NULL == pdevice)
   {
      MPPLNX_VHBA_BUG();
   }
   /*
   * get the lock first
   */
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   pHostEntry= mppLnxVHBAContext.phostList;
   if(pHostEntry == NULL)
   {
       MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
          "mppLnx_get_PdeviceEntry: did not have phostList\n"));
      return NULL;
   }
   
   for(listPrt = mppLnxVHBAContext.phostList->list.next; 
         listPrt != &(mppLnxVHBAContext.phostList->list); 
		 listPrt =listPrt->next)
   {
      pHostEntry = list_entry(listPrt, mppLnx_PhostEntry_t, list);
      if(pHostEntry->hostNo == hostNo)
      {
         break;
      }
	  else
      {
         pHostEntry = NULL;
      }
   }
   /*
   * the device is from a new host
   */
   if(pHostEntry == NULL)
   {
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "mppLnx_get_PdeviceEntry: did not have host entry\n"));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return NULL;
   }
   /*
   * The device is from new target
   */
   if(pHostEntry->targetList == NULL)
   {
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "mppLnx_get_PdeviceEntry: did not have targetList\n"));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return NULL;
   }
   /*
   * find the target for this device
   */
   for(listPrt = pHostEntry->targetList->list.next; 
         listPrt != &(pHostEntry->targetList->list); 
		 listPrt =listPrt->next)
   {
      pTargetEntry = list_entry(listPrt, mppLnx_PtargetEntry_t, list);
      if(pTargetEntry->targetId == targetId)
      {
         break;
      }else
      {
         pTargetEntry = NULL;
      }
   }
   /*
   * the device is from a new target
   */
   if(pTargetEntry == NULL)
   {
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
         "mppLnx_get_PdeviceEntry: did not have target entry\n"));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return NULL;
   }
   
   pDeviceEntry =  pTargetEntry->pdevices[lun];
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
      pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun)); 
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   return pDeviceEntry;
 
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_insert_PdeviceEntry
* SUMMARY:  creates and insert a mppLnx_PDeviceEntry_t object to mppLnx_VHNAContext
* SCOPE:   public
*
* DESCRIPTION:
   This function creates a "mppLnx_PdeviceEntry_t object" based on the input 
   arguments and inserts the object to the MPP virtual HBA driver's 
   internal data model.
   If an object mppLnx_PdeviceEntry_t, which represents the input "pdevice",  already 
   exists in the internal data model, this function doesn't insert the object to 
   the list again.

* @pdevice A physical Scsi_Device object
* @rdacInfo the RdacDeviceInformation_t object that represents the arrary where the physical
            device is discovered
* @controllerSlot The controller slot number from where the physcial device is discovered
* @pathId   The path ID number of the physical device in its LunPathObjects
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_insert_PdeviceEntry
(
   Scsi_Device * pdevice, 
   RdacDeviceInformation_t *rdacInfo, 
   int controllerSlot, 
   int pathId
   )
{
   int                        size = 0;
   int                        hostNo = 0;
   int                        targetId =0;
   int                        lun = 0;
   struct list_head           * listPrt;
   mppLnx_PhostEntry_t        * pHostEntry = NULL;
   mppLnx_PtargetEntry_t      * pTargetEntry = NULL;
   unsigned long              iflag;

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
      "Entering mppLnx_insert_PdeviceEntry()\n"));

   hostNo = pdevice->host->host_no;
   targetId = pdevice->id;
   lun = pdevice->lun;
   /*
   * get the lock first
   */
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   /*check if the phostList is initialized
   */
   if(mppLnxVHBAContext.phostList == NULL)
   {
      /*
      * create an empty entry 
      * this entry is only used as list_head
      */
      size = sizeof(mppLnx_PhostEntry_t);
      mppLnxVHBAContext.phostList = kmalloc(size, GFP_ATOMIC);
      if(mppLnxVHBAContext.phostList == NULL)
      {
         /* error case*/
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 421,
            "Can not get kernel memory for allacating mppLnx_PhostEntry_t\n"));
         spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
         return;
      }
      else
      {
         MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
            "create an empty mppLnx_PhostEntry_t as list_head\n"));
      }

      memset(mppLnxVHBAContext.phostList, 0, size);
      
      INIT_LIST_HEAD(&(mppLnxVHBAContext.phostList->list));
  
   }

   /*
   * find its host
   */
   for(listPrt = mppLnxVHBAContext.phostList->list.next; 
         listPrt != &(mppLnxVHBAContext.phostList->list); 
		 listPrt =listPrt->next)
   {
      pHostEntry = list_entry(listPrt, mppLnx_PhostEntry_t, list);
      if(pHostEntry->hostNo == hostNo)
      {
         MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
            "found the mppLnx_PhostEntry_t for the device host_no %d\n",
            hostNo));
         break;
      }else
      {
         pHostEntry = NULL;
      }
   }
   /*
   * the device is from a new host
   */
   if(pHostEntry == NULL)
   {
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
         "Did not found the mppLnx_PhostEntry_t for the device host_no %d\n",
            hostNo));
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
         "Create a new mppLnx_PhostEntry_t\n"));
      pHostEntry = mppLnx_createPhostEntry(pdevice);
      list_add_tail(&(pHostEntry->list), &(mppLnxVHBAContext.phostList->list));
   }
   /*
   * check if its target is created
   */
   if(pHostEntry->targetList == NULL)
   {
      /**
      * there is no mppLnx_PtargetEntry object in the host's target list
      * create an empty one and it will be the list_head
      */
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
         " create a list_head for mppLnx_PtargetEntry_t \n"));
      size = sizeof(mppLnx_PtargetEntry_t);
      pHostEntry->targetList = kmalloc(size,GFP_ATOMIC);
      if(NULL == pHostEntry->targetList)
      {
         /* error case. Cannot get kernel memory*/
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 422,
            "Cannot get kernel memory for allacating mppLnx_PtargetEntry_t\n"));
         spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
         return;
      }
      memset(pHostEntry->targetList, 0, size);

      INIT_LIST_HEAD(&(pHostEntry->targetList->list));
   }
   /*
   * find its target
   */
   for(listPrt = pHostEntry->targetList->list.next; 
         listPrt != &(pHostEntry->targetList->list);
		 listPrt =listPrt->next)
   {
      pTargetEntry = list_entry(listPrt, mppLnx_PtargetEntry_t, list);
      if(pTargetEntry->targetId == targetId)
      {
         break;
      }else
      {
         pTargetEntry = NULL;
      }
   }
   /*
   *From a new target
   */
   if(NULL == pTargetEntry)
   {
      /*
      * did not find the target entry for this device
      * create one
      */
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
         "create an mppLnx_PtargetEntry_t for device h%dc%dt%dl%d\n",
         pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun));
      pTargetEntry = mppLnx_createPtargetEntry(pdevice);
      if(NULL == pTargetEntry)
      {
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 423,
            "Error in geting memory for mppLnx_PtargetEntry_t\n"));
         spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
         return;
      }
      list_add_tail(&(pTargetEntry->list),&(pHostEntry->targetList->list));
      size = sizeof(mppLnx_PDeviceEntry_t *)* mppLnx_Config_Parameters.mpp_MaxLunsPerArray;
      pTargetEntry->pdevices = 
			(mppLnx_PDeviceEntry_t * *) kmalloc(size, GFP_ATOMIC);
      if(pTargetEntry->pdevices == NULL)
      {
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 424,
            "Cannot allocate memory for mppLnx_PDeviceEntry_t **)\n"));
         spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
         return;
      }
      memset(pTargetEntry->pdevices, 0, size);
   }
   
   /*
   * Check whether or not the physical device entry is alreay created
   */
   if(pTargetEntry->pdevices[lun] != NULL)
   {
      MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
            "The mppLnx_PDeviceEntry is already in the list\n"));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return;
   }

   /* 
   * get the target entry. Create mppLnx_PDeviceEntry_t
   */
   size = sizeof(mppLnx_PDeviceEntry_t);
   pTargetEntry->pdevices[lun] = kmalloc(size, GFP_ATOMIC);

   if(pTargetEntry->pdevices[lun] == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 425,
            "Cannot allocate memory for mppLnx_PDeviceEntry_t)\n"));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return;
   }

   memset(pTargetEntry->pdevices[lun], 0 ,size);
   pTargetEntry->pdevices[lun]->pdevice = pdevice;
   pTargetEntry->pdevices[lun]->rdacInfo = rdacInfo;
   pTargetEntry->pdevices[lun]->controllerSlot = controllerSlot;
   pTargetEntry->pdevices[lun]->pathId = pathId;
   /*
   * create a list_header for the current command list
   */
   size = sizeof(mppLnx_ProxyRequestEntry_t);
   pTargetEntry->pdevices[lun]->currentCmndList = kmalloc(size, GFP_ATOMIC);

   if(pTargetEntry->pdevices[lun]->currentCmndList == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 426,
            "Cannot allocate memory for mppLnx_ProxyRequestEntry_t list header)\n"));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return;
   }

   INIT_LIST_HEAD(&(pTargetEntry->pdevices[lun]->currentCmndList->list));
   mppLnx_createPDeviceProcNode(pTargetEntry->pdevices[lun]);
   
   /*
   * release the lock
   */
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_3,
      "Exiting mppLnx_insert_PdeviceEntry()\n"));
   
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_delete_PdeviceEntry
* SUMMARY:  remove a mppLnx_PDeviceEntry_t object from mppLnx_VHNAContext,
            delete its proc node, and free its memory
* SCOPE:   public
*
* DESCRIPTION:
   This function remove a "mppLnx_PDeviceEntry_t object", based on the input
   Scsi physical device, from the MPP virtual HBA driver's internal data
   model.
   It removes this mppLnx_PDeviceEntry_t object's /proc file and frees this 
   object's memory.
*  @pdevice a physical Scsi_Device object to be deleted.

* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*       This function performs a reverse operation as the function of
*       mppLnx_insert_PdeviceEntry does
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_delete_PdeviceEntry(Scsi_Device * pdevice)
{
   int                        hostNo = 0;
   int                        targetId =0;
   int                        lun = 0;
   struct list_head *         listPrt;
   mppLnx_PhostEntry_t        * pHostEntry = NULL;
   mppLnx_PtargetEntry_t      * pTargetEntry = NULL;
   mppLnx_PDeviceEntry_t      * pDeviceEntry = NULL;
   unsigned long              iflag;
  
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_delete_PdeviceEntry() h%dc%dt%dl%d\n",
      pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun));
   
   hostNo = pdevice->host->host_no;
   targetId = pdevice->id;
   lun = pdevice->lun;
 
   /*
   * get the lock first
   */
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   pHostEntry = mppLnxVHBAContext.phostList;

   if(NULL == pHostEntry)
   {
      /* why no Host device
      * the phsyical HBA driver must not be loaded
      */
     MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
        "progaming error:pHostEntry= mppLnxVHBAContext.phostList is NULL\n"));
     MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
        "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
              pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun));
     spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
     return -1;
   }
   
   
   
   for(listPrt = mppLnxVHBAContext.phostList->list.next; 
         listPrt != &(mppLnxVHBAContext.phostList->list); 
		 listPrt =listPrt->next)
   {
      pHostEntry = list_entry(listPrt, mppLnx_PhostEntry_t, list);
      if(pHostEntry->hostNo == hostNo)
      {
         MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
            "mppLnx_delete_PdeviceEntry found the device's host\n"));
         break;
      }
	  else
      {
         pHostEntry = NULL;
      }
   }
   
   if(pHostEntry == NULL)
   {
      /* why no physical host
      * the phsyical HBA driver must not be loaded.
      * the upper level driver messed up?
      */
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "mppLnx_delete_PdeviceEntry did not have the device's host\n"));
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
              pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun));
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return -1;
   }
   
   if(pHostEntry->targetList == NULL)
   {
      /* why no target
      * the upper level driver messed up?
      */
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "mppLnx_delete_PdeviceEntry: null targetLis\n"));
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
         pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun)); 
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return -1;
   }
   /*
   *
   */
   for(listPrt = pHostEntry->targetList->list.next; 
         listPrt != &(pHostEntry->targetList->list); 
		 listPrt =listPrt->next)
   {
      pTargetEntry = list_entry(listPrt, mppLnx_PtargetEntry_t, list);
      if(pTargetEntry->targetId == targetId)
      {
         MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
            "mppLnx_delete_PdeviceEntry: found the device's target object\n"));
         break;
      }else
      {
         pTargetEntry = NULL;
      }
   }
   
   if(pTargetEntry == NULL)
   {
      /* why no target
      * the upper level driver messed up?
      */
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "mppLnx_delete_PdeviceEntry: cannot find the device's target object\n"));
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
         pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun)); 
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return -1;
   }
   
   pDeviceEntry =  pTargetEntry->pdevices[lun];
   if(NULL == pDeviceEntry)
   {
      /* why no the physical device
      */
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "mppLnx_delete_PdeviceEntry: cannot find the device's mppLnx_PDeviceEntry_t\n"));
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
         pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun)); 
      spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
      return -1;
   }

   /**
   * clean its proc node and free its memory
   */
   mppLnx_removePDeviceProcNode(pDeviceEntry);
   
   /**
   * what do we do if there are proxy commands still going on. This case will not
   * happen. This is because the middle level driver checks where or not the Scsi_Device
   * objects, which are provided by this host, are still used by upper level drivers.
   * The module can be unloaded only when no upper level drivers are using it
   */
   kfree(pDeviceEntry->currentCmndList); /* free the empty list head */

   kfree(pDeviceEntry);
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_get_PdeviceEntry() h%dc%dt%dl%d\n",
      pdevice->host->host_no, pdevice->channel, pdevice->id, pdevice->lun)); 
   return 0;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_createPtargetEntry
* SUMMARY:  Creates and initializes an mppLnx_PtargetEntry_t object.
* SCOPE:   local
*
* DESCRIPTION:
   This function creates and initializes an mppLnx_PtargetEntry_t object. 
   The object is not linked/added to its list.
* @ pdevice A physical Scsi_Device object for which its mppLnx_PtargetEntry_t
   object needs to be created
* RESTRICTIONS:
* 
* RETURNS: An mppLnx_PtargetEntry_t object which is newly created or NULL if we
*           have memory allocation problem
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_PtargetEntry_t * mppLnx_createPtargetEntry(Scsi_Device * pdevice)
{
	int                        size = 0;
   mppLnx_PtargetEntry_t      * ptargetEntry;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_createPtargetEntry()\n"));
     /*
      * create an entry 
      */
      size = sizeof(mppLnx_PtargetEntry_t);
      ptargetEntry = kmalloc(size, GFP_ATOMIC);
      
      if(ptargetEntry == NULL)
      {
         /* error. Cannot get kernel memory */
         MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 427,
            "Cannot get kernel memory for mppLnx_PtargetEntry_t\n"));
         return NULL;
      }
      memset((void *) ptargetEntry, 0, size);
      ptargetEntry->targetId = pdevice->id;
   
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_createPtargetEntry()\n"));
   return ptargetEntry;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_createVDeviceEntry
* SUMMARY:  Creates and initializes an mppLnx_VDeviceEntry_t object.
* SCOPE:   local
*
* DESCRIPTION:
   This function creates and initializes an mppLnx_VDeviceEntry_t object. The object 
   is not linked/added to its list.

* @ lun The Lun number of the virtual Scsi_Device
* @rdacInfo A RdacDeviceInformat_t object which represents a storage array,
* RESTRICTIONS:
* 
* RETURNS: A newly created mppLnx_VDeviceEntry_t object or NULL if memory allocation 
            error.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_VDeviceEntry_t * mppLnx_createVDeviceEntry(
    int lun,
    RdacDeviceInformation_t *rdacInfo)
{
   int                        size = 0;
   mppLnx_VDeviceEntry_t      * vDeviceEntry;
   
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_createVDeviceEntry()\n"));
   /*
    * create an entry 
   */
   size = sizeof(mppLnx_VDeviceEntry_t);
   vDeviceEntry = kmalloc(size, GFP_ATOMIC);
      
   if(vDeviceEntry == NULL)
   {
      /* error cannot get kernel memory */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 428,
           "Cannot get kernel memory for mppLnx_VDeviceEntry_t\n"));
      return NULL;
   }
   memset((void *) vDeviceEntry, 0, size);
   vDeviceEntry->lun = lun;
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
       "Exiting mppLnx_createVDeviceEntry()\n"));
   return vDeviceEntry;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_createProxyRequest
* SUMMARY:  Creates and initializes an mppLnx_ProxyRequestEntry_tobject.
* SCOPE:   local
*
* DESCRIPTION:
   This function creates and initializes an mppLnx_ProxyRequestEntry_t object. 
   The object is not linked/added to its list.

* @ vscp The command which  received from a virtual device
* @proxyCmnd The proxy Scsi_Cmnd object which is going to issue to a physical
             HBA driver
* 
* RETURNS: An mppLnx_ProxyRequestEntry_t object or NULL if memory allocation error
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_ProxyRequestEntry_t * mppLnx_createProxyRequest(
      Scsi_Cmnd * vscp,
      Scsi_Cmnd * proxyCmnd)
{
   mppLnx_ProxyRequestEntry_t    * request;
   int                           size = sizeof(mppLnx_ProxyRequestEntry_t);

    MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_createProxyRequest\n"));
   request =  kmalloc(size,GFP_ATOMIC);
   if(NULL == request)
   {
      /* cannot get memory */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 429,
         "Error cannot get kernel memory for mppLnx_ProxyRequestEntry_t\n"));
      return NULL;
   }

   memset(request, 0 , size);
   request->vcmnd = vscp;
   request->pcmnd = proxyCmnd;
   request->serial_number = (unsigned long) proxyCmnd;
   MPP_GETTIME(request->startTime);
   MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_createProxyRequest\n"));
   return request;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_insert_proxyRequest
* SUMMARY:  Inserts a proxy request entry to its physical device's request list
* SCOPE:    public
*
* DESCRIPTION:
   This function inserts mppLnx_ProxyRequestEntry_t object the virtual HBA's 
   internal data model. Based on the information in "*request", this function 
   will look up to the internal data model and insert the "request"  to its 
   mppLnx_PdeviceEntry_t object's currentCmndList.

*  @request A mppLnx_ProxyReqyestEntry_t object which represents the association of
*           a virtual Scsi command and its proxy Scsi command.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_insert_proxyRequest(mppLnx_ProxyRequestEntry_t *request)
{
   Scsi_Device                   *pdevice;
   mppLnx_PDeviceEntry_t         * pde;
   unsigned long                 iflag;

   MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_inser_proxyRequest() v SN %ld p cmnd 0x%lx\n",
      request->vcmnd->serial_number, (unsigned long)request->pcmnd));

   pdevice = request->pcmnd->device;
   
   pde = mppLnx_get_PdeviceEntry(pdevice);
   if(NULL == pde)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 430,
         "Error: did not find the p device's mppLnx_PDeviceEntry_t\n"));
      return;
   }
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   list_add_tail(&(request->list), &(pde->currentCmndList->list));
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_inser_proxyRequest\n"));
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_get_virtualRequest
* SUMMARY:  Query virtual Scsi_Cmnd based on a proxy Scsi_Cmnd.
* SCOPE:    public
*
* DESCRIPTION:
   This function is used to look up a mppLnx_ProxyRequesetEntry_t object based on 
   a proxy Scsi command. This function is called every time when a proxy command is 
   completed. The returned mppLnx_ProxyRequestEntry_t object is used to complete the
   proxy command's virtual Scsi command.
   This function will remove the mppLnx_ProxyRequesetEntry_t object from the MPP 
   virtual HBA driver's internal data model.

* @proxyCmnd the proxy Scsi_Cmnd for which its virtual Scsi_Cmnd is searching.
*
* RESTRICTIONS:
* 
* RETURNS:Returns the mppLnx_ProxyRequestEntry_t object that is associated with 
            the "*proxyCmnd" or null if the look-up is failed.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_ProxyRequestEntry_t * mppLnx_get_virtualRequest(Scsi_Cmnd *proxyCmnd)
{
   struct list_head                 * listPt;
   mppLnx_ProxyRequestEntry_t       * foundRv = NULL;
   mppLnx_PDeviceEntry_t            *pde;
   Scsi_Device                      *pdevice;
   unsigned long                    iflag;
   
   MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_get_virtualRequest() \n"));
   pdevice = proxyCmnd->device;

   pde = mppLnx_get_PdeviceEntry(pdevice);
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   for(listPt = pde->currentCmndList->list.next;
            listPt != &(pde->currentCmndList->list);
			listPt = listPt->next)
   {
      foundRv = list_entry(listPt, mppLnx_ProxyRequestEntry_t, list);
      if(foundRv->serial_number == (unsigned long)proxyCmnd)
      {
         MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
            "Found the proxy request for the proxyCmnd h%dc%dt%dl%d\n",
            proxyCmnd->device->host->host_no,
            proxyCmnd->channel, proxyCmnd->target, proxyCmnd->lun));
         MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
            " p command serial_number 0x%lx command %x v cmnd SN %ld\n",
            foundRv->serial_number, proxyCmnd->cmnd[0],foundRv->vcmnd->serial_number));
         break;
      }else
      {
         foundRv = NULL;
      }
   }

   if(foundRv == NULL)
   {
      /* 
      * program error. Why the proxy request is not in our internal data model
      * this is a big problem, panic the kernel*/

      MPPLNX_VHBA_BUG();
   }
   else
   {
      list_del(&(foundRv->list));
   }
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   MPP_DEBUGPRINT((MPP_READ_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_get_virtualRequest() v SN %ld p cmnd 0x%lx\n",
      foundRv->vcmnd->serial_number,(unsigned long)proxyCmnd));
   return foundRv;

}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_getAnyPathByVDev
* SUMMARY:  Gets a virtual devive's physical device
* SCOPE:   local
*
*
* DESCRIPTION:
   This function gets a physical path of a virtual scsi device.

* @ vsdp the input argument is a virtual Scsi_Device object for whom its
      physical path to be queried
* RESTRICTIONS:
* 
* RETURNS: any available physical path of the virtual device
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
Scsi_Device * mppLnx_getAnyPathByVDev(Scsi_Device * vsdp)
{
   Scsi_Device                   *psdp = NULL;
   RdacDeviceInformation_t       *rdacInfo;
   RdacLunInformation_t          *rdacLun;
   int                           path = 0;
   int                           controller = 0;
   
   rdacInfo = mpp_ModuleArray[vsdp->id].RdacInfo;
   if( rdacInfo == NULL )
   {
	MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
            "mppLnx_getAnyPathByVDev()- rdacInfo is NULL.\n" ));

	return NULL;
   }

   rdacLun  = & rdacInfo->RdacLuns[vsdp->lun];
   for(controller = 0; controller < 2; controller++)
   {
      for(path = 0; path < mppLnx_Config_Parameters.mpp_MaxPathsPerController; path++)
      {
         psdp = rdacLun->LunControllers[controller]->LunPaths[path].LunPathDevice;
         if(psdp != NULL)
         {
            break;
         }else
         {
            psdp = NULL;
         }
      }

      if(psdp != NULL)
      {
         break;
      }
   }
   return psdp;
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_update_PdeviceStat
* SUMMARY:  Updates a mppLnx_PDeviceEntry_t object's statistics.
* SCOPE:   local
*
*
* DESCRIPTION:
   This function updates a mppLnx_PDeviceEntry_t object's statistics information.
   This function is called after an IO is completed via a physical device.

* @ vde The mppLnx_PDeviceEntry_t object whose statistics needs to be updated
* @ioStatus The completion status of an IO
* @roundTrip the time spent on the IO in the unit of Linux jiffies.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_update_PdeviceStat(
        mppLnx_PDeviceEntry_t *pde,
        int ioStatus, 
        unsigned long roundTrip)
{
   unsigned long        iflag;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_update_PdeviceStat()\n"));
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   switch (ioStatus)
   {
      case MPPLNX_IO_SUCCESS:
         pde->ioCount++;
         if(pde->shortestTrip == 0)
         {
            pde->shortestTrip = roundTrip;
         }else if(roundTrip < pde->shortestTrip)
         {
            pde->shortestTrip = roundTrip;
         }

         if(roundTrip > pde->longestTrip)
         {
            pde->longestTrip = roundTrip;
         }
         break;
      case MPPLNX_IO_PATHFAILOVER:
         pde->numPathFailures++;
         break;
      case MPPLNX_IO_FAILED:
         pde->ioFails++;
         break;
      case MPPLNX_IO_CTRLFAILOVER:
      default:
         /* 
         *The physical device should not receive CTRLFAILOVER IO state
         */
         printk("Unsupported IO completion state\n");
         break;
   }
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_update_PdeviceStat()\n"));
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_update_VdeviceStat
* SUMMARY:  update a mppLnx_VDeviceEntry_t object's statistics
* SCOPE:   local
*
*
* DESCRIPTION:
   Updates a mppLnx_VDeviceEntry_t object's statistics information.

* @vde a mppLnx_VDeviceEntry_t object whose statistics information to be updated.
* @ioStatus The status of an IO to the virtual device. Value can be MPPLNX_IO_SUCESS
      MPPLNX_IO_FAILED, MPPLNX_IO_PATHFAILOVER or MPPLNX_IO_CTRLFAILOVER
  @roundTrip the time spent on the IO in the unit of Linux jiffies.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_update_VdeviceStat(
       mppLnx_VDeviceEntry_t * vde, 
       int ioStatus, 
       unsigned long roundTrip)
{
   unsigned long        iflag;
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_update_VdeviceStat()\n"));
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   switch (ioStatus)
   {
      case MPPLNX_IO_SUCCESS:
         vde->ioCount++;
         if(vde->shortestTrip == 0)
         {
            vde->shortestTrip = roundTrip;
         }else if(roundTrip < vde->shortestTrip)
         {
            vde->shortestTrip = roundTrip;
         }

         if(roundTrip > vde->longestTrip)
         {
            vde->longestTrip = roundTrip;
         }
         break;
      case MPPLNX_IO_PATHFAILOVER:
         vde->numPathFailures++;
         break;
      case MPPLNX_IO_FAILED:
         vde->ioFails++;
         break;
      case MPPLNX_IO_CTRLFAILOVER:
         vde->numCntlFailures++;
         if(vde->shortestCntlFailoverTime == 0)
         {
            vde->shortestCntlFailoverTime = roundTrip;
         }
         else if(roundTrip < vde->shortestCntlFailoverTime)
         {
            vde->shortestCntlFailoverTime = roundTrip;
         }

         if(roundTrip > vde->longestCntlFailoverTime)
         {
            vde->longestCntlFailoverTime = roundTrip;
         }

         break;
      default:
         /* 
         *The physical device should not receive CTRLFAILOVER IO state
         */
         printk("Unsupported IO completion state %d\n",ioStatus);
         break;
   }
   
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_update_VdeviceStat()\n"));

}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_createPhostEntry
* SUMMARY:  Creates and initializes an mppLnx_PhostEntry_t object.
* SCOPE:   local
*
*
* DESCRIPTION:
   This function creates and initializes mppLnx_PhostEntry_t object. The object
   is not linked/added to its list.

* @pdevice A physical Scsi_Device object whose mppLnx_PhostEntry_t object needs
   to be created.

* RESTRICTIONS:
* 
* RETURNS: A newly created mppLnx_PhostEntry_t object.
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
mppLnx_PhostEntry_t * mppLnx_createPhostEntry(Scsi_Device * pdevice)
{
   mppLnx_PhostEntry_t     * phostEntry;
   int                     size = 0;
   
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_createPhostEntry()\n"));
   /*
   * create an entry 
   */
   size = sizeof(mppLnx_PhostEntry_t);
   phostEntry = kmalloc(size, GFP_ATOMIC);
      
   if(phostEntry == NULL)
   {
      /* error in getting memory
      */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL, 431,
            "Cannot get kernel memory for mppLnx_PhostEntry_t\n"));
      return NULL;
   }
   memset((void *) phostEntry, 0, size);
   phostEntry->host = pdevice->host;
   phostEntry->hostNo = pdevice->host->host_no;
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_createPhostEntry()\n"));
   return phostEntry;
}
