/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************
NAME            mppLnx_vhbaproc.c
SUMMARY         %description%
VERSION         %version: 7 %
UPDATE DATE     %date_modified: Tue Nov 18 17:34:27 2003 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:This source file contains Linux MPP virtual HBA driver functions 
   that are related to the creation/deletion/management/information-output of 
   the device topology tree. In the Linux MPP environment, the device topology 
   tree is implemented in the Linux /proc file system. The functions in this file
   manages the creation of the leaf node of the topology tree. The Linux MPP upper
   level driver manages the directory nodes of the device topology tree.


INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:
  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

#define __SRCmppLnx_vhbaproc_c

/***  INCLUDES  ***/  
/***  Linux kernel INCLUDES  ***/  
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/stat.h>
#include <asm/io.h>

#include <linux/blk.h>
/*
* Linux SCSI middle level include
*/
#include "scsi.h"
#include "hosts.h"
#include "sd.h"


/***  MPP commomn API INCLUDES  ***/  
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "MPP_Sysdep.h"
#include "mppLnx_version.h"

/***  Linux MPP Virtual HBA INCLUDES  ***/  
#include "mppLnx_vhba.h"
#include "mppLnx_shared.h"


extern Scsi_Host_Template * driver_template;
extern mppLnx_VHBAContext_t mppLnxVHBAContext;
/***  CONSTANT DEFINITIONS  ***/
#define MPPLNX_VIRTUALLUN_PREFIX       "virtualLun"
#define MPPLNX_PHYSICALLUN_PREFIX      "LUN"
#define MPPLNX_UTMLUN_PREFIX           "UTM_"
#define MPPLNX_VLUN_READ_BUFFER_SIZE   2048

/***  MACRO DEFINITIONS  ***/


/***  TYPE DEFINITIONS  ***/


/***  LOCALS  ***/


/***  PROCEDURES  ***/



/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_createVDeviceProcNode
* SUMMARY:  Creates a virtual Scsi_Device's proc node
* SCOPE:    public
   
*
* DESCRIPTION:
   This function creates a virtual device /proc leaf node. It is invoked at the 
   time of building the virtual HBA driver internal data model by the mppLnx_detect() 
   function. 
   This function uses Linux /proc file system APIs  "proc_mkdir()" and 
   "create_proc_entry()" to create virtual device proc leaf node. The proc entry's 
   read_proc function pointer will point to mppLnx_Vdevice_read_proc() .
* 
*  @vde the virtual device entry object for which the /proc node needs to be created
* RESTRICTIONS:
*     This function can not be invoked from an interrupt handler since proc entry creation
*     uses the memory with GFP_KERNEL flag
* 
* RETURNS: SUCCESS indicates the proc leaf node is successfully 
*           created or FAILED otherwise
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

int mppLnx_createVDeviceProcNode(mppLnx_VDeviceEntry_t * vde)
{
   int                        rv = FAILED;
   struct proc_dir_entry      * parent = NULL;
   RdacDeviceInformation_t    *rdacInfo = NULL;

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_createVDeviceProcNode()\n"));
   
   if(NULL == vde)
   {
      MPPLNX_VHBA_BUG();
   }

   rdacInfo = vde->rdacInfo;

   if(NULL == rdacInfo)
   {
      MPPLNX_VHBA_BUG();
   }

   parent = rdacInfo->ModuleHandle;

   if(NULL == parent)
   {
      MPPLNX_VHBA_BUG();
   }

   vde->procNode.parent = parent;
   sprintf(vde->procNode.nodeName,"%s%d",
      MPPLNX_VIRTUALLUN_PREFIX,
      vde->lun);
   vde->procNode.procMode =  PROC_LEAF_NODE_MODE;
  
   vde->procNode.thisNode = create_proc_read_entry(
         vde->procNode.nodeName,
         vde->procNode.procMode,
         vde->procNode.parent,
         mppLnx_VDevice_read_proc, 
         vde
         );
   
   if(vde->procNode.thisNode == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,470,
         "Not enough memory to reigster MPP virtual HBA in /proc/scsi/mpp !\n"));
      rv = FAILED;
   }else
   {
      vde->procNode.thisNode->owner = THIS_MODULE;
      rv = 0;
      vde->procNode.isCreated = 1;
      vde->procNode.needsClean = 0;
      vde->procNode.nodeType = MPPLNX_VDEVICE_NODE;
      rv = SUCCESS;
   }
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exitering mppLnx_createVDeviceProcNode() for virtual device T%dL%d\n",
      vde->rdacInfo->VirtualTargetId,
      vde->lun));
   return rv;

}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_createPDeviceProcNode
* SUMMARY:  Creates physical scsi device's /proc node
* SCOPE:   local
*
*
* DESCRIPTION:
   This function creates a physical device /proc leaf node. It is invoked at the time 
   of building the virtual HBA driver internal data model by the mppLnx_detect() function. 
   This function uses Linux /proc file system APIs  "proc_mkdir()" and "create_proc_entry()"
   to create virtual device proc leaf node. The proc entry's read_proc function pointer 
   will point to mppLnx_Pdevice_read_proc()
* @pde the physical device entry object for which the /proc node needs to be created.
*
* RESTRICTIONS:RICTIONS:
*     This function can not be invoked from an interrupt handler since proc entry creation
*     uses the memory with GFP_KERNEL flag
* 
* RETURNS:SUCCESS indicates the proc leaf node is successfully 
*           created or FAILED otherwise
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_createPDeviceProcNode(mppLnx_PDeviceEntry_t * pde)
{
   RdacDeviceInformation_t       *rdacInfo =NULL;
   struct proc_dir_entry         *parent = NULL;
   int                           rv = FAILED;
   int                           ControllerIndex;
   int                           PathIndex;

   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_createPDeviceProcNode()\n"));
   rdacInfo = pde->rdacInfo;
   ControllerIndex = pde->controllerSlot;
   PathIndex = pde->pathId;
   parent = rdacInfo->ControllerInfo[ControllerIndex]->ControllerPath[PathIndex].DirectoryVertex;
   if(NULL == parent)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,471,
         "the physical device does not have a parent array %s controller %d path %d\n",
         (char *) rdacInfo->ModuleName, ControllerIndex, PathIndex));
      MPPLNX_VHBA_BUG();
   }

   pde->procNode.parent = parent;
   
   /*
   * check if the device is UTM device. If so, its proc node name will be
   * UTM_LUNn
   */
   if(!rdacInfo->RdacLuns[pde->pdevice->lun].UTMLun)
   {
      sprintf(pde->procNode.nodeName,"%s%d",
         MPPLNX_PHYSICALLUN_PREFIX,
         pde->pdevice->lun);
   }else
   {
      sprintf(pde->procNode.nodeName,"%s%s%d",
         MPPLNX_UTMLUN_PREFIX,
         MPPLNX_PHYSICALLUN_PREFIX,
         pde->pdevice->lun);
   }

   pde->procNode.procMode =  PROC_LEAF_NODE_MODE;
   pde->procNode.thisNode = create_proc_read_entry(
         pde->procNode.nodeName,
         pde->procNode.procMode,
         pde->procNode.parent,
         mppLnx_PDevice_read_proc, 
		   (void *)pde
         );
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Made PDevice Proc Node\n"));
   
 
   if(pde->procNode.thisNode == NULL)
   {
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,472,
         "Not enough memory to reigster MPP virtual HBA in /proc/scsi/mpp !\n"));
      rv = FAILED;
   }else
   {
      rv = SUCCESS;
      pde->procNode.isCreated = 1;
      pde->procNode.needsClean = 0;
      pde->procNode.nodeType = MPPLNX_PDEVICE_NODE;
      pde->procNode.thisNode->owner = THIS_MODULE;
   }
   MPP_DEBUGPRINT((MPP_INIT_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_createPDeviceProcNode()\n"));
   return rv;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_removeVDeviceProcNode
* SUMMARY:  Removes virtual device /proc node.
* SCOPE:     local
*
*
* DESCRIPTION:
   This function removes a virtual device /proc leaf node. It is invoked at the time
   of the virtual HBA driver unloading time by the mppLnx_release() function. 
   This function uses Linux /proc file system APIs  "remove_proc_entry()" to remove
   virtual device proc leaf node. 
* @vde the virtual device entry for which its /proc node will be removed
* RESTRICTIONS:
* 
* RETURNS:SUCCESS indicates the proc leaf node is successfully 
*           removed or FAILED otherwise
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_removeVDeviceProcNode(mppLnx_VDeviceEntry_t * vde)
{
  

   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_removeVDeviceProcNode()\n"));
   remove_proc_entry(vde->procNode.nodeName, vde->procNode.parent);
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exitng mppLnx_removeVDeviceProcNode()\n"));

   
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_removePDeviceProcNode
* SUMMARY:  Removes a physical device's /proc node
* SCOPE: local
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This function removes a physical device /proc leaf node. It is invoked at the 
   time of the virtual HBA driver unloading time by the mppLnx_release() function. 
   This function uses Linux /proc file system APIs  "remove_proc_entry()" to remove 
   virtual device proc leaf node. 
   
*  @pde the physical device entry for which its /proc node will be removed
* RESTRICTIONS:
* 
* RETURNS:SUCCESS indicates the proc leaf node is successfully 
*           removed or FAILED otherwise
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_removePDeviceProcNode(mppLnx_PDeviceEntry_t * pde)
{
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_removePDeviceProcNode()\n"));
   remove_proc_entry(pde->procNode.nodeName, pde->procNode.parent);
   MPP_DEBUGPRINT((MPP_DETACH_DEBUG+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_removePDeviceProcNode()\n"));
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_VDevice_read_proc
* SUMMARY:  Wrtie a virtual device's info to the user space via /proc file system
* SCOPE:    public 
*
*
* DESCRIPTION:
   This function is assigned to /proc file system's proc_dir_entry 
   data structure's (*read_proc) function pointer at a virtual device /proc leaf
   node creation time. The "void *data" is an object of "mppLnx_VdeviceEntry_t" 
   object. This function is invoked when an end user types 
   "cat  /proc/scsi/mpp/arrayName/virtualLun#".
   
*  @page the pointer of write-buffer. 
*  @start 
*  @offset the function should start writing at an offset of "offset"   
*  @count  at most "count" bytes that the function can write
*  @eof  write 1 to indicate End Of File
*  @data  driver specific data pointer. It points to an mppLnx_VDeviceEntry_t object
*           this function
* RESTRICTIONS:
* 
* RETURNS: This function returns the number of bytes written into the "page
*
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_VDevice_read_proc
(
      char *page, 
      char ** start, 
      off_t offset, 
      int count, 
      int *eof, 
      void *data
      )
{
   mppLnx_VDeviceEntry_t      *vde =(mppLnx_VDeviceEntry_t *) data;
   int                        len = 0;
   int                        position = 0;
   int                        begin;
   char                       asciiWWN[34];
   RdacDeviceInformation_t    * rdacInfo = vde->rdacInfo;
   RdacLunInformation_t       * lunInfo = &(rdacInfo->RdacLuns[vde->lun]);
   static char                vDeviceReadBuffer[MPPLNX_VLUN_READ_BUFFER_SIZE];

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_VDevice_read_proc()\n"));
   memset(vDeviceReadBuffer,0, MPPLNX_VLUN_READ_BUFFER_SIZE);

   len = sprintf(&vDeviceReadBuffer[position],"Linux MPP driver. Version:%s Build:%s\n",
         MPPLNX_VERSION,MPPLNX_BUILDTIMESTAMP);

   position = position +len;
   mpp_ConvertWWNtoAscii(asciiWWN,&lunInfo->WWN[0]);
   len = sprintf(&vDeviceReadBuffer[position],"Lun WWN:%s\n",asciiWWN);

   if(vde->vdevice == NULL)
   {
      position = position +len;
      len = sprintf(&vDeviceReadBuffer[position],
         "The virtual device for array %s Lun %d is not available\n",
         (char *)rdacInfo->ModuleName, vde->lun);
   }
   else
   {
      position = position +len;
      len = sprintf(&vDeviceReadBuffer[position],
         "Virtual Scsi Address: host_no:%d channel:%d target:%d Lun:%d\n",
         vde->vdevice->host->host_no,
         vde->vdevice->channel,
         vde->vdevice->id,
         vde->vdevice->lun);
   }

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "I/O Statistics:\n");

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tNumber of IOs:%d\n",vde->ioCount);

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tLongest trip of all I/Os:%d\n",vde->longestTrip);

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tShortest trip of all I/Os:%d\n",vde->shortestTrip);

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tNumber of occurrences of path failover events:%d\n",vde->numPathFailures);

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tNumber of occurrences of controller failover events:%d\n",
      vde->numCntlFailures);

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tThe longest Controller Failover Time:%ld\n",
      vde->longestCntlFailoverTime);

   position = position +len;
   len = sprintf(&vDeviceReadBuffer[position],
      "\tThe shortest Controller Failover Time:%ld\n",
      vde->shortestCntlFailoverTime);

   position = position +len;

   /*
   * let's copy the buffer to the page
   */
   if(position < offset)
   {
      len = 0;
      begin = position;
   }
   else
   {
      begin = offset;
      len = position - offset;
      if(len > count)
      {
         len = count;
      }
   }

   if(len < count)
   {
      (*eof) = 1;
   }else
   {
      (*eof) = 0;
   }
   
   if(len > 0)
   {
      memcpy(page+offset, &vDeviceReadBuffer[begin], len);
   }

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_VDevice_read_proc()\n"));
   return len;


}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_PDevice_read_proc
* SUMMARY:  Wrtie a physical device's info to the user space via /proc file system
* SCOPE:    public 
*
* DESCRIPTION:
   This function is assigned to /proc file system's proc_dir_entry 
   data structure's (*read_proc) function pointer at a physical device /proc 
   leaf node creation time. The "void *data" is an object of "mppLnx_PdeviceEntry_t" 
   object. This function is invoked when an end user types 
   "cat  /proc/scsi/mpp/arrayName/Controller#/path#/Lun#".

   
* @page the pointer of write-buffer. 
*  @start
*  @offset    
*  @count
*  @eof
*  @data  driver specific data pointer. It points to an mppLnx_PDeviceEntry_t object
*           this function
* RESTRICTIONS:
* 
* RETURNS:This function returns the number of bytes written into the "page".
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_PDevice_read_proc
(
      char *page, 
      char ** start, 
      off_t offset, 
      int count, 
      int *eof, 
      void *data
      )
{
   mppLnx_PDeviceEntry_t         *pde =(mppLnx_PDeviceEntry_t *) data;
   int                           len = 0;
   int                           position = 0;
   int                           begin;
   char                          asciiWWN[34];
   RdacDeviceInformation_t       * rdacInfo = pde->rdacInfo;
   RdacLunInformation_t          * lunInfo = &(rdacInfo->RdacLuns[pde->pdevice->lun]);
   struct list_head              * listPt;
   mppLnx_ProxyRequestEntry_t    * proxyRequest;
   static char                   pDeviceReadBuffer[MPPLNX_PROC_READ_BUFFSIZE];
   MPP_TIME                      currentTime;
   unsigned long                 iflag;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Entering mppLnx_PDevice_read_proc()\n"));
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Input arguments offset %d count %d \n",
       offset, count));
   memset(pDeviceReadBuffer,0, MPPLNX_PROC_READ_BUFFSIZE);

   len = sprintf(&pDeviceReadBuffer[position],"Linux MPP driver. Version:%s Build:%s\n",
         MPPLNX_VERSION,MPPLNX_BUILDTIMESTAMP);

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "wrote version info to the buffer\n"));
   position = position +len;
   if(NULL == lunInfo)
   {
      /* programming error 
      */
      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "Cannot get WWN from array %s lun %d\n",
         (char *) rdacInfo->ModuleName, pde->pdevice->lun));
      MPPLNX_VHBA_BUG();
      return 0;
   }
   mpp_ConvertWWNtoAscii(asciiWWN,&lunInfo->WWN[0]);
   len = sprintf(&pDeviceReadBuffer[position],"Lun WWN:%s\n",asciiWWN);
   
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "put wwn to the buufer\n"));
   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "Physical HBA driver: %s\n",
      pde->pdevice->host->hostt->proc_name);
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "put host proc_name to the buffer\n"));

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "Device Scsi Address: host_no:%d channel:%d target:%d Lun:%d\n",
      pde->pdevice->host->host_no,
      pde->pdevice->channel,
      pde->pdevice->id,
      pde->pdevice->lun);
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "put device ID to the buufer\n"));

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "I/O Statistics:\n");

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "\tNumber of IOs:%d\n",pde->ioCount);

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "\tLongest trip of all I/Os:%d\n",pde->longestTrip);

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "\tShortest trip of all I/Os:%d\n",pde->shortestTrip);

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "\tNumber of occurences of IO failed events:%d\n",pde->ioFails);

   position = position +len;
   len = sprintf(&pDeviceReadBuffer[position],
      "Outstanding IOs on this device:\n");
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "put I/O statics to the buffer\n"));

   /** 
   * walk through the list of currentCmndList and report all of them
   */
   spin_lock_irqsave(mppLnxVHBAContext.vhbaContextLock,iflag);
   if(pde->currentCmndList->list.next == &(pde->currentCmndList->list))
   {
      /* there is no outstading proxy command*/

      MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
         "No outstanding proxy commands in the list\n"));
      position = position +len;
      len = sprintf(&pDeviceReadBuffer[position],"\t\tNo outstanding commands\n");
   }
   else
   {
      for(listPt = pde->currentCmndList->list.next;
         listPt != (&pde->currentCmndList->list);
         listPt = listPt->next )
      {
         MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
            "get currentCmnd \n"));
         proxyRequest = list_entry(listPt, mppLnx_ProxyRequestEntry_t, list);
      
         position = position +len;
         len = sprintf(&pDeviceReadBuffer[position],"\t\tCommand: %02x:\n",
            proxyRequest->vcmnd->cmnd[0]);

         position = position +len;
         len = sprintf(&pDeviceReadBuffer[position],
            "\t\tvirtual cmnd: serial %ld, proxy cmnd address: 0x%lx\n",
            proxyRequest->vcmnd->serial_number,
            proxyRequest->serial_number);

         position = position +len;
         MPP_GETTIME(currentTime);
         len = sprintf(&pDeviceReadBuffer[position],
            "\t\tHow long it has been queued:%ld\n",
            currentTime - proxyRequest->startTime);
      }
   }
   spin_unlock_irqrestore(mppLnxVHBAContext.vhbaContextLock,iflag);

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "put outstanding I/O to the buffer\n"));
   
   position = position +len;

   /*
   * let's copy the buffer to the page
   */
   if(position < offset)
   {
      len = 0;
      begin = position;
   }
   else
   {
      begin = offset;
      len = position - offset;
      if(len > count)
      {
         len = count;
      }
   }

   if(len < count)
   {
      (*eof) = 1;
   }else
   {
      (*eof) = 0;
   }
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "put outstanding I/O to the buufer\n"));
   
   if(len >0 )
   {
      memcpy(page+offset, &pDeviceReadBuffer[begin], len);
   }

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_PDevice_read_proc()\n"));
   return len;

}
