/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************


NAME            mppLnx_vhbatask.c
SUMMARY         %description%
VERSION         %version: 23.1.2 %
UPDATE DATE     %date_modified: Thu Feb 26 12:52:55 2004 %
PROGRAMMER      %created_by:    yqi %

        Copyright 2003, LSI Logic Storage Systems, Inc.

DESCRIPTION:
   This source file contains function that are related to Linux MPP virtual HBA 
   driver's kernel thread management and tasklet management. The kernel threads 
   include:
   1. The fail back scan kernel thread. The fail back scan kernel thread wakes-up 
   in a specified time period and performs fail back scan function that is provided 
   by the MPP common code
   2.The sync IO kernel thread. The sync IO kernel thread performs sync IOs that 
   are generated during the virtual HBA's IO routing. Currently, the sync IOs are 
   those IOs related to the SCSI-2 reserve/release translation layer

   The tasklet is used for performing delayed IOs and dispatching proxy command to 
   the Linux SCSI middle level.


INCLUDE FILES:

NOTES:

RESTRICTIONS:

SEE ALSO:

REFERENCE:
  349-1033500 Linux MPP Driver Feature Architecture Specification
  349-1034610 Linux MPP Virtual HBA Driver Feature Design Document
  349-1034750 Linux MPP Upper Level Driver Feature Design Document
  349-1034730 Linux MPP Utility Feature Design Document
  349-1034620 Linux MPP Installation Feature Design Document

IMPLEMENTATION:

MODIFICATION HISTORY:

*******************************************************************************/

#define __SRCmppLnx_vhbatask_c

/***  INCLUDES  ***/  
/***  Linux kernel INCLUDES  ***/  
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <asm/io.h>

#include <linux/blk.h>
#include "scsi.h"
#include "hosts.h"
#include "sd.h"

#include <linux/stat.h>

/***  MPP commomn API INCLUDES  ***/  
#include "MPP_Common.h"
#include "MPP_RdacInfo.h"
#include "MPP_ProtoTypes.h"
#include "MPP_Sysdep.h"
#include "mppLnx_version.h"
#include "mppCmn_s2tos3.h"

/***  Linux MPP Virtual HBA INCLUDES  ***/  
#include "mppLnx_vhba.h"
#include "mppLnx_shared.h"



/***  MACRO DEFINITIONS  ***/
#define MPPLNX_FAILBACK_PROC_NAME           "mppFailback"
#define MPPLNX_SYNCIO_WORKER_PROC_NAME      "mppWorker"


/***  TYPE DEFINITIONS  ***/


/***  LOCALS  ***/
extern void (*middle_level_timeoutfunc)(Scsi_Cmnd *);
#if USE_THREAD_FOR_DPC
extern mppLnx_dpcQueueContext_t   mppLnxDpcQueueContext;
#else
extern mppLnx_taskletQueue_t mppLnxTaskletQueue;
#endif
extern spinlock_t taskQueue_lock;
extern mppLnx_failbackScanContext_t mppLnxFailbackScanContext;
extern mppLnx_workerQueueContext_t mppLnxWorkerQueueContextContext;
/***  PROCEDURES  ***/




/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_failback_handler
* SUMMARY:  The entry function of the virtual HBA driver failback scan thread
* SCOPE:    public 
*
*
* DESCRIPTION:
   This function is the failback rescan kernel thread entry function. The input
   argument "void *data" is not used. This function performs the following operations:
      The process name of this kernel thread in the process table will be "mpp_failback".
      This kernel thread will be demonized so that it has no associated control terminal 
         and its stdin, stdout and stderr are closed.
      Its parent process id be init's process id (1). 
      If the virtual HBA driver is loaded as a module, this kernel thread ignores 
         all signals except SIGHUP that is used for system shutdown and module unloading.
         Signal SIGHUP will cause the failback kernel thread terminated.
      Starts an infinity loop and uses Linux kernel timer and Linux semaphore to schedule 
         failback rescan.
*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

int mppLnx_failback_handler(void * data)
{
   struct Scsi_Host     *host = (struct Scsi_Host *) data;
   unsigned long        currentTime;

   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_failback_handler\n"));

   /*
    * We only listen to signals if the virtual HBA was loaded as a module.
    * If the virtual HBA was compiled into the kernel, then we don't listen
    * to any signals.
    */
   if( host->loaded_as_module ) 
   {
	   siginitsetinv(&current->blocked, SIGHUP);
	} 
   else 
   {
	   siginitsetinv(&current->blocked, 0);
   }
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_failback_handler Locks the kernel\n"));
	lock_kernel();

	/*
	 * make it a daemon process with ppid 1
	 */
	daemonize();
	reparent_to_init();

	/*
	 * Set the name of this process and initialize its internal data fields
	 */
	sprintf(current->comm, MPPLNX_FAILBACK_PROC_NAME);
	mppLnxFailbackScanContext.failback_process   = current;
   
   /*
   *set the time-out function
   */
   mppLnxFailbackScanContext.fbs_timer.function = mppLnx_rescan_timer_handler;
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_failback_handler release the kernel\n"));

	unlock_kernel();

	/*
	 * Wake up the thread that created us.
	 */
	MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_failback_handler--Wake up parent from failback scan\n"));

	up(mppLnxFailbackScanContext.detect_sem);
   /*
   * make the expires less than jiffies for the first time
   */
   mppLnxFailbackScanContext.fbs_timer.expires = jiffies -10;
   atomic_set(&mppLnxFailbackScanContext.scan_active,0);

	while (1) {
		/*
		 * If we get a signal, it means we are supposed to go
		 * away and die.  This typically happens if the user is
		 * trying to unload a module.
		 */
		MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
         "failback scan sleeping\n"));
      /**
      * add timer. This thread will wake up after the timer goes off
      * If the timer is still alive, don't add it again. Just modify
      * the expires.
      */
      currentTime = jiffies;
      if(currentTime < mppLnxFailbackScanContext.fbs_timer.expires)
      {
         mppLnxFailbackScanContext.fbs_timer.data = (unsigned long)NULL;
         mppLnxFailbackScanContext.fbs_timer.expires = jiffies + 
               mppLnx_Config_Parameters.mpp_ScanInterval*HZ;
         MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
         "mppLnx_failback_handler -- mod rescan timer\n"));
         mod_timer(&mppLnxFailbackScanContext.fbs_timer,
            jiffies + mppLnx_Config_Parameters.mpp_ScanInterval*HZ);
      }
      else
      {
         mppLnxFailbackScanContext.fbs_timer.data = (unsigned long)NULL;
         mppLnxFailbackScanContext.fbs_timer.expires = jiffies + 
               mppLnx_Config_Parameters.mpp_ScanInterval*HZ;
         MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_failback_handler -- add rescan timer\n"));
         add_timer(&mppLnxFailbackScanContext.fbs_timer);
      }


      /*
       * Note - we always use down_interruptible with the semaphore
       * even if the module was loaded as part of the kernel.  The
       * reason is that down() will cause this thread to be counted
       * in the load average as a running process, and down
       * interruptible doesn't.  Given that we need to allow this
       * thread to die if the driver was loaded as a module, using
       * semaphores isn't unreasonable.
       */
      down_interruptible(mppLnxFailbackScanContext.failback_sem);
      
      if( host->loaded_as_module ) 
      {
         if (signal_pending(current))
            break;
      }

      MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
         "Failback scan handler waking up\n"));
      /*
      * perform the fail back scan only if the scanning is not active
      */
      if(atomic_read(&mppLnxFailbackScanContext.scan_active) == 0)
      {
        atomic_set(&mppLnxFailbackScanContext.scan_active,1);
        mpp_FailBackScan();
        mpp_CheckAVTState();
        mpp_CheckPreferredAVTOwner(FALSE);
        atomic_set(&mppLnxFailbackScanContext.scan_active,0);
        MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
         "Failback scan handler-- performed failbackScan\n"));
         
      }else
      {
         MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
           "Failback is still running\n"));
      }

   } /*while (1) */

   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Failback scan exiting\n"));

   /*
    * Make sure that nobody tries to wake us up again.
    */
   mppLnxFailbackScanContext.failback_sem       = NULL;
   mppLnxFailbackScanContext.failback_process   = NULL;
   atomic_set(&mppLnxFailbackScanContext.scan_active, 0);
   /*
   * delete the timer
   */
   del_timer(&(mppLnxFailbackScanContext.fbs_timer));
   /*
    * If anyone is waiting for us to exit (i.e. someone trying to unload
    * a driver), then wake up that process to let them know we are on
    * the way out the door.  This may be overkill - I *think* that we
    * could probably just unload the driver and send the signal, and when
    * the error handling thread wakes up that it would just exit without
    * needing to touch any memory associated with the driver itself.
    */
   if (mppLnxFailbackScanContext.detect_sem != NULL)
   {
      up(mppLnxFailbackScanContext.detect_sem);
      mppLnxFailbackScanContext.detect_sem = NULL;
   }
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_failback_handler\n"));

   return 0;

}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_worker_handler
* SUMMARY:  The entry function of the virtual HBA driver sync IO worker thread
* SCOPE:    public 
*
*
* DESCRIPTION:
  This function is the sync IO worker thread entry function. The input
   argument "void *data" is not used. This function performs the following operations:
      The process name of this kernel thread in the process table will be "mppWorker".
      This kernel thread will be demonized so that it has no associated control terminal 
         and its stdin, stdout and stderr are closed.
      Its parent process id be init's process id (1). 
      If the virtual HBA driver is loaded as a module, this kernel thread ignores 
         all signals except SIGHUP that is used for system shutdown and module unloading.
         Signal SIGHUP will cause the failback kernel thread terminated.
      Starts an infinity loop, sleeps on a semaphore and wakes up when a sysn command 
      needs to be processed
*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
int mppLnx_worker_handler(void * data)
{
   struct Scsi_Host              *host = (struct Scsi_Host *) data;
   mppLnx_CmndEntry_t            * cmndEntry = NULL;
   struct list_head              * listPt;
   unsigned long                 iflages;
   RdacDeviceInformation_t       *rdacInfo;
   unsigned char                 reservationKey[8];
   LWORD                         mpp_status;
   Scsi_Cmnd                     *vcmnd;
   void                          (*vcmnd_timeoutfunc)(Scsi_Cmnd *);
   SenseData_t                   sensedata;
   TINY                          dataDirection;

   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_worker_handler\n"));

   /*
    * We only listen to signals if the HA was loaded as a module.
    * If the HA was compiled into the kernel, then we don't listen
    * to any signals.
    */
   if( host->loaded_as_module ) 
   {
	   siginitsetinv(&current->blocked, SIGHUP);
	} 
   else 
   {
	   siginitsetinv(&current->blocked, 0);
   }
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_worker_handler() lock the kernel\n"));
	lock_kernel();

	/*
	 * make it a daemon process with ppid 1
	 */
	daemonize();
	reparent_to_init();

	/*
	 * Set the name of this process.
	 */
	sprintf(current->comm, MPPLNX_SYNCIO_WORKER_PROC_NAME);
   mppLnxWorkerQueueContextContext.workerQueue_process   = current;
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_worker_handler released the kernel\n"));

	unlock_kernel();

	/*
	 * Wake up the thread that created us.
	 */
	MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "mppLnx_worked_handler--Wake up parent from syncIO thread\n"));

	up(mppLnxWorkerQueueContextContext.detect_sem);

   /*
   * the thread is not busy
   */
   atomic_set(&mppLnxWorkerQueueContextContext.worker_active,0);

	while (1) {
      /*
      * perform our real work here
      */
      /*
      * lock the queue and pick up one cmnd entry at a time. Remove the cmnd entry
      * from the list and unlock. In this way, we don't block any new entry being
      * added to the list.
      */
      cmndEntry = NULL;
      spin_lock_irqsave(mppLnxWorkerQueueContextContext.cmndQueueLock, iflages);
      listPt = mppLnxWorkerQueueContextContext.cmndEntry->list.next;
      if(listPt != &(mppLnxWorkerQueueContextContext.cmndEntry->list))
      {
         cmndEntry = list_entry(listPt, mppLnx_CmndEntry_t, list);
         list_del(&cmndEntry->list);

      }
      spin_unlock_irqrestore(mppLnxWorkerQueueContextContext.cmndQueueLock, iflages);

      /*
      * process the command
      */
      if(cmndEntry != NULL)
      {
         /*
         *mark the thread is busy
         */
         atomic_set(&mppLnxWorkerQueueContextContext.worker_active,1);
         vcmnd = cmndEntry->vcmnd;
         /*
         * delete the timer for this command
         */
         vcmnd_timeoutfunc = (void (*)(Scsi_Cmnd *))cmndEntry->vcmnd_timeoutfunc;
         MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
            "before delete timer vcmnd 0x%lx SNno 0x%lx \n",
            (unsigned long)vcmnd, vcmnd->serial_number));

         switch(vcmnd->cmnd[0])
         {
            case RESERVE:
            case RESERVE_10:
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "received Reserve command\n"));

               mppCmn_GenReservationKey(mppLnx_Config_Parameters.mpp_S2ToS3Key
                  ,reservationKey);
               rdacInfo = mpp_ModuleArray[vcmnd->target].RdacInfo;
               mpp_status = mppCmn_MapScsi2Reserve(rdacInfo,
                  vcmnd->lun,
                  mppLnx_Config_Parameters.mpp_S2ToS3Key,
                  reservationKey,
                  TRUE,
                  &sensedata);
               
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "mppCmn_MapScsi2Reserve returned mpp_status %d\n",mpp_status));

               if(mpp_status == MPPCMN_PRTN_GOOD_STATUS)
               {
                  vcmnd->result = 0x00;
               }
               else if(mpp_status == MPPCMN_PRTN_RESERVATION_CONFLICT)
               {
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,460,
                     "mppVhba reserve failure - RESERVATION CONFLICT %d:%d:%d:%d\n",
                     vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));

                  vcmnd->result = (RESERVATION_CONFLICT <<1);
               }
               else if(mpp_status == MPPCMN_PRTN_CHECK_CONDITION)
               {
                  /*
                  * 1 log the check condition
                  * 2 re-wrap the check condition to 02/04/09
                  * 3. complete the command with check-condition
                  */
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,468,
                     "mppVhba reserve failure - Check-Condition %d:%d:%d:%d\n",
                     vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));
                  mppLnx_setCheckCondition(vcmnd,&sensedata);

               }else
               {
                  /* return an error. It will be translated to selection time-out*/
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,461,
                     "mppVhba reservation failure - Cannot talk to the deivice %d:%d:%d:%d\n",
                     vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));
                  vcmnd->result = (DID_NO_CONNECT << 16);
               }

               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "Ready to add timer back vcmnd 0x%lx, timeout 0x%x timeout_func 0x%lx\n",
                  (unsigned long)vcmnd,
                  vcmnd->timeout_per_command,
                  (unsigned long)vcmnd_timeoutfunc));
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "before add timer vcmnd 0x%lx SNno 0x%lx \n",
                     (unsigned long)vcmnd, vcmnd->serial_number));

               scsi_add_timer(vcmnd,5*HZ,vcmnd_timeoutfunc);

               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "before call scsi_done done 0x%lx \n",
                  (unsigned long)vcmnd->scsi_done));

               vcmnd->scsi_done(vcmnd);

               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "completed Reserve command\n"));
               break;
            case RELEASE:
            case RELEASE_10:
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "received Release command\n"));

               mppCmn_GenReservationKey(mppLnx_Config_Parameters.mpp_S2ToS3Key
                  ,reservationKey);
               rdacInfo = mpp_ModuleArray[vcmnd->target].RdacInfo;
               mpp_status = mppCmn_MapScsi2Release(rdacInfo,
                  vcmnd->lun,
                  mppLnx_Config_Parameters.mpp_S2ToS3Key,
                  reservationKey,
                  FALSE,
                  &sensedata);

               if(mpp_status == MPPCMN_PRTN_GOOD_STATUS)
               {
                  vcmnd->result = 0x00;
               }
               else if(mpp_status == MPPCMN_PRTN_RESERVATION_CONFLICT)
               {
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,465,
                     "S2toS3 translation : Mapping release gets a RESERVATION_CONFLICT return-code\n"));
                  vcmnd->result = (RESERVATION_CONFLICT <<1);
               }
               else if(mpp_status == MPPCMN_PRTN_CHECK_CONDITION)
               {
                  /*
                  * 1 log the check condition
                  * 2 re-wrap the check condition to 02/04/09
                  * 3. complete the command with check-condition
                  */
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,469,
                     "mppVhba release failure - Check-Condition %d:%d:%d:%d\n",
                     vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));
                  mppLnx_setCheckCondition(vcmnd,&sensedata);

               }else
               {
                  /* return an error. It will be translated to selection time-out*/
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,462,
                     "mppVhba release failure - Cannot talk to the deivice %d:%d:%d:%d\n",
                     vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));
                  vcmnd->result = (DID_NO_CONNECT << 16);
               }
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "Ready to add timer back vcmnd 0x%lx, timeout 0x%x timeout_func 0x%lx\n",
                  (unsigned long)vcmnd,
                  vcmnd->timeout_per_command,
                  (unsigned long)vcmnd_timeoutfunc));
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "before add timer vcmnd 0x%lx SNno 0x%lx \n",
                     (unsigned long)vcmnd, vcmnd->serial_number));

               scsi_add_timer(vcmnd,5*HZ,vcmnd_timeoutfunc);
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "before call scsi_done\n"));
               vcmnd->scsi_done(vcmnd);
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "completed Release command\n"));
               break;
            case MPPLNX_BDR_CMND_OPCODE:
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "received bus device reset  command\n"));
               rdacInfo = mpp_ModuleArray[vcmnd->target].RdacInfo;
               mpp_status = mppCmn_MapBusDeviceReset(rdacInfo,vcmnd->lun);
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "completed bus device reset command. status:%d\n",mpp_status));
               /*
               * the virtual command is a copy of the orignal command from the middle
               * level. We need to free the virtual command
               */
               kfree(vcmnd);
               /*
               * The bus device reset message is not a real command. We don't need to 
               * complete it. The scsi middle level has a dummy scsi_done function.
               */
               break;

            default:
               rdacInfo = mpp_ModuleArray[vcmnd->target].RdacInfo;
               if(vcmnd->sc_data_direction==SCSI_DATA_READ)
               {
                  dataDirection = MPPCMN_XFER_DEV_HOST;
               }
               else if(vcmnd->sc_data_direction==SCSI_DATA_WRITE)
               {
                  dataDirection = MPPCMN_XFER_HOST_DEV;
               }
               else if(vcmnd->sc_data_direction==SCSI_DATA_NONE)
               {
                  dataDirection = MPPCMN_XFER_NONE;
               }
               else
               {
                  dataDirection = MPPCMN_XFER_UNKNOWN;
               }

               mpp_status = mppCmn_CheckForReservationHolder(rdacInfo,
                  vcmnd->lun,
                  TRUE,
                  &sensedata);

               if(mpp_status == MPPCMN_PRTN_GOOD_STATUS)
               {
                  /*
                  * Either this node is the persistent reservation holder of this LUN or the LUN 
                  * is not reserved. Let's dispatch the special command as a normal command. 
                  * Note we break the "switch" to avoid the command being completed.
                  */
                  mppLnx_schedule_queuecmd(vcmnd);
                  break;
               }
               else if(mpp_status == MPPCMN_PRTN_RESERVATION_CONFLICT)
               {
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,470,
                     "mppVhba command %x failure - RESERVATION CONFLICT %d:%d:%d:%d\n",
                     vcmnd->cmnd[0],vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));

                  vcmnd->result = (RESERVATION_CONFLICT <<1);
               }else if(mpp_status == MPPCMN_PRTN_CHECK_CONDITION)
               {
                  /*
                  * 1 log the check condition
                  * 2 re-wrap the check condition to 02/04/09
                  * 3. complete the command with check-condition
                  */
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,471,
                     "mppVhba command %x failure - Check-Condition %d:%d:%d:%d\n",
                     vcmnd->cmnd[0],vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));

                  mppLnx_setCheckCondition(vcmnd,&sensedata);

               }else
               {
                  /* return an error. It will be translated to selection time-out*/
                  MPP_ERRORLOGPRINT((MPP_ERR_FATAL,461,
                     "mppVhba command %x failure - Cannot talk to the deivice %d:%d:%d:%d\n",
                     vcmnd->cmnd[0],vcmnd->host->host_no,vcmnd->channel,vcmnd->target,vcmnd->lun));
                  vcmnd->result = (DID_NO_CONNECT << 16);
               }
               scsi_add_timer(vcmnd,5*HZ,vcmnd_timeoutfunc);
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "before call scsi_done\n"));
               vcmnd->scsi_done(vcmnd);
               MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
                  "completed command %x\n",vcmnd->cmnd[0]));
               break;
         }
         /*
         * free the cmndEntry
         */
         kfree(cmndEntry);
      }
      else
      {
         /*
		   * If we get a signal, it means we are supposed to go
		   * away and die.  This typically happens if the user is
		   * trying to unload a module.
		   */
		  MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
           "mpp worker thread sleeping\n"));
        /*
        * before going to sleep, mark the thread not-busy
        */
        atomic_set(&mppLnxWorkerQueueContextContext.worker_active,0);

         /*
		   * Note - we always use down_interruptible with the semaphore
		   * even if the module was loaded as part of the kernel.  The
		   * reason is that down() will cause this thread to be counted
		   * in the load average as a running process, and down
		   * interruptible doesn't.  Given that we need to allow this
		   * thread to die if the driver was loaded as a module, using
		   * semaphores isn't unreasonable.
		   */
		   down_interruptible(mppLnxWorkerQueueContextContext.workerQueue_sem);
         if( host->loaded_as_module ) 
         {
			   if (signal_pending(current))
				   break;
         }

		   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
            "SyncIO worker handler waking up\n"));
      }

	}
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "syncIO worker thread exiting\n"));

	/*
	 * Make sure that nobody tries to wake us up again.
	 */
	mppLnxWorkerQueueContextContext.workerQueue_sem = NULL;
   /*
   * set the process to NULL. The mppLnx_release needs to check
   * the process value before senging a sig
   */
   mppLnxWorkerQueueContextContext.workerQueue_process = NULL;
	/*
   * free memory and resources
   */
   
	/*
	 * If anyone is waiting for us to exit (i.e. someone trying to unload
	 * a driver), then wake up that process to let them know we are on
	 * the way out the door.  This may be overkill - I *think* that we
	 * could probably just unload the driver and send the signal, and when
	 * the error handling thread wakes up that it would just exit without
	 * needing to touch any memory associated with the driver itself.
	 */
	if (mppLnxWorkerQueueContextContext.detect_sem != NULL)
   {
		up(mppLnxWorkerQueueContextContext.detect_sem);
      mppLnxWorkerQueueContextContext.detect_sem = NULL;
   }
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_worker_handler\n"));

	return 0;
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_rescan_timer_handler
* SUMMARY:  The failback scan thread's timer-callback function
* SCOPE:    local
*
*
* DESCRIPTION:
   This is the failback rescan thread's timer callback function. This function 
   will be assigned to rescan timer's timeout callback function "(*function)". 
   The input data is the address of "mppLnx_sem", which is a Linux semaphore.
   This function simply calls Linux kernel semaphore API  "up()" to raise the 
   semaphore.

*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_rescan_timer_handler(unsigned long data)
{
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_rescan_timer_handler\n"));
   /*we only wake up the failback snan thread if it is sleep
    */
   if(atomic_read(&mppLnxFailbackScanContext.scan_active) == 0)
   {
      MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
         "mppLnx_rescan_timer_handler -- wake up failback scan thread\n"));
      up(mppLnxFailbackScanContext.failback_sem);
   }
   else
   {
      MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
         "mppLnx_rescan_timer_handler -- failback scan is in prograss\n"));
   }
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_rescan_timer_handler.\n"));
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_failbackRescan
* SUMMARY:  Scan failback one.
* SCOPE:    public 
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This function is called by the Linux MPP upper level driver after receiving 
   an failback rescan IOCTL from mppUtil program.
   This function simply calls Linux kernel semaphore API  "up()" to raise 
   the rescan semaphore. That will cause the rescan kernel thread to perform 
   failback rescan.

*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

void mppLnx_failbackRescan()
{
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Enterting mppLnx_failbackRescan()\n"));
   mppLnx_rescan_timer_handler((unsigned long) NULL);
   MPP_DEBUGPRINT((MPP_FAILOVERSCAN_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_failbakcRescan()\n"));
}


#if USE_THREAD_FOR_DPC

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_dpc_handler
* SUMMARY:  The entry function of the virtual HBA driver command dispatching/commpletion/retires
 thread
* SCOPE:    public 
*
*
* DESCRIPTION:
   This function is the command dispatch/completion/retries entry function. The input
   argument "void *data" is the virtual host. 
   This function performs the following operations:
      The process name of this kernel thread in the process table will be "mpp_dcr".
      This kernel thread will be demonized so that it has no associated control terminal 
         and its stdin, stdout and stderr are closed.
      Its parent process id be init's process id (1). 
      If the virtual HBA driver is loaded as a module, this kernel thread ignores 
         all signals except SIGHUP that is used for system shutdown and module unloading.
         Signal SIGHUP will cause the DCR kernel thread terminated.
      Starts an infinity loop and uses Linux kernel timer and Linux semaphore to schedule 
         its dispatching, completion and retries tasks. The thread will be waked up by
        1. a new virtual command is added to command queue for dispatching
        2. a proxy command is completed and added to command queue for 
*
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/

int mppLnx_dpc_handler(void * data)
{
   struct Scsi_Host     *host = (struct Scsi_Host *) data;
   unsigned long        currentTime;
   mppLnx_dpcQueueContext_t         * taskQueuep = &mppLnxDpcQueueContext;
   mppLnx_CmndEntry_t            * cmndEntry;
   struct list_head              * listPt;
   unsigned long                 iflages;
   MPP_TIME                      ControllerIoWaitTime;
   MPP_TIME                      ArrayIoWaitTime;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Entering mppLnx_dpc_handler\n"));

   /*
    * We only listen to signals if the virtual HBA was loaded as a module.
    * If the virtual HBA was compiled into the kernel, then we don't listen
    * to any signals.
    */
   if( host->loaded_as_module ) 
   {
	   siginitsetinv(&current->blocked, SIGHUP);
	} 
   else 
   {
	   siginitsetinv(&current->blocked, 0);
   }
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "mppLnx_dpc_handler Locks the kernel\n"));
	lock_kernel();

	/*
	 * make it a daemon process with ppid 1
	 */
	daemonize();
	reparent_to_init();

	/*
	 * Set the name of this process and initialize its internal data fields
	 */
	sprintf(current->comm, "mpp_dcr");
	mppLnxDpcQueueContext.dpcQueue_process   = current;
   
	unlock_kernel();

	/*
	 * Wake up the thread that created us.
	 */
	MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "mppLnx_dpc_handler--Wake up parent from failback scan\n"));

	up(mppLnxDpcQueueContext.detect_sem);

   /*
   * Set the command queue to need schedule
   */
   atomic_set(&mppLnxDpcQueueContext.needSchedule,1);

   /*
   * The infinity loop. It can only break by SIGHUP
   */
	while (1) 
   {
      /*
      *Loop through all commands in the command queue. It breaks when the queue is empty
      */
      while(1)
      {
         atomic_set(&mppLnxDpcQueueContext.needSchedule,0);
         spin_lock_irqsave(mppLnxDpcQueueContext.cmndQueueLock, iflages);
         if( (listPt = taskQueuep->cmndEntry->list.next)
           != &(taskQueuep->cmndEntry->list) )
         {
            cmndEntry = list_entry(listPt, mppLnx_CmndEntry_t, list);
            spin_unlock_irqrestore(mppLnxDpcQueueContext.cmndQueueLock, iflages);
            if(cmndEntry->taksType == MPPLNX_DELAY_RESPONSE_TYPE)
            {
               /*
               * completing delayed response command 
               */
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "Completing scsi delayed virtual scsi_cmnd vtarget %d vlun %d\n",
                  cmndEntry->vcmnd->target, cmndEntry->vcmnd->lun));
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "the virtual command serial_number %ld command %x\n",
                  cmndEntry->vcmnd->serial_number, cmndEntry->vcmnd->cmnd[0]));
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "vcmnd->result %d\n", cmndEntry->vcmnd->result));
               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "mppLnx_tasklet_func cmnd %lx serial_number %ld\n",
                  (unsigned long)cmndEntry->vcmnd, cmndEntry->vcmnd->serial_number));
               cmndEntry->vcmnd->scsi_done(cmndEntry->vcmnd);
            }
            else if(cmndEntry->taksType == MPPLNX_QUEUED_CMND_TYPE)
            {
               /*dispatch the proxy command to the physical HBA
               */
               MPP_GETTIME(currentTime);
               MPP_GETTIME(ControllerIoWaitTime);
               MPP_GETTIME(ArrayIoWaitTime);

               MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
                  "queue the command to the physical HBA driver cmnd %02x %02x\n",
                  cmndEntry->vcmnd->cmnd[0],
                  cmndEntry->vcmnd->cmnd[1]));

               mppLnx_do_queuecommand(cmndEntry->vcmnd,
                  middle_level_timeoutfunc,
                  cmndEntry->vcmnd->scsi_done,
                  UNKNOWN_PREVIOUS_CONTROLLER,
                  UNKNOWN_PREVIOUS_PATH,
                  0, /*CommandTimeoutRetryCount*/
                  mppLnx_Config_Parameters.mpp_SyncRetryCount, /*RetryCount */
                  0, /*SelectionTimeoutRetryCount*/
                  0, /*UaRetryCount*/
                  currentTime,   /*the current time */
                  ControllerIoWaitTime, /*controller io wait start time*/
                  ArrayIoWaitTime   /*the array io wait start time */
                  );
            }
#if USE_THREAD_FOR_RETRY
            else if(cmndEntry->taksType == MPPLNX_COMPLETED_CMND_TYPE)
            {
                __mppLnx_scsi_done(cmndEntry->vcmnd);
            }
#endif
            else
            {
               /*
               * we should not get here
               */
               MPPLNX_VHBA_BUG();
            }
            /*remove the entry from the list*/
            spin_lock_irqsave(mppLnxDpcQueueContext.cmndQueueLock, iflages);
            list_del(&cmndEntry->list);
            spin_unlock_irqrestore(mppLnxDpcQueueContext.cmndQueueLock, iflages);
            kfree(cmndEntry);
         }else
         {
            spin_unlock_irqrestore(mppLnxDpcQueueContext.cmndQueueLock, iflages);
            atomic_set(&taskQueuep->needSchedule,1);
            break;
            
         }
      } /*the command loop */
      /*
      * If we get a signal, it means we are supposed to go
      * away and die.  This typically happens if the user is
      * trying to unload a module.
		*/
      MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "mpp_dpc thread sleeping\n"));
      /*
      * Note - we always use down_interruptible with the semaphore
      * even if the module was loaded as part of the kernel.  The
      * reason is that down() will cause this thread to be counted
      * in the load average as a running process, and down
      * interruptible doesn't.  Given that we need to allow this
      * thread to die if the driver was loaded as a module, using
      * semaphores isn't unreasonable.
      */
		down_interruptible(mppLnxDpcQueueContext.dpc_sem);
      if( host->loaded_as_module ) 
      {
         if (signal_pending(current))
            break;
      }

		MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
            "dpc handler waking up\n"));

   } /*while (1) */

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "dpc thread exiting\n"));

   /*
    * Make sure that nobody tries to wake us up again.
    */
   mppLnxDpcQueueContext.dpc_sem       = NULL;
   mppLnxDpcQueueContext.dpcQueue_process   = NULL;
   atomic_set(&mppLnxDpcQueueContext.needSchedule, 0);

   /*
    * If anyone is waiting for us to exit (i.e. someone trying to unload
    * a driver), then wake up that process to let them know we are on
    * the way out the door.  This may be overkill - I *think* that we
    * could probably just unload the driver and send the signal, and when
    * the error handling thread wakes up that it would just exit without
    * needing to touch any memory associated with the driver itself.
    */
   if (mppLnxDpcQueueContext.detect_sem != NULL)
   {
      up(mppLnxDpcQueueContext.detect_sem);
      mppLnxDpcQueueContext.detect_sem = NULL;
   }
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting dpc_handler\n"));

   return 0;

}
#else
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_tasklet_func
* SUMMARY:  Process deffered virtual Scsi_Cmnd when the virtual HBA's tasklet 
*           is scheduled to run
* SCOPE:    public 
*
*
* DESCRIPTION:
   This function is called every time when the virtual HBA driver's tasklet 
   is scheduled to run. This function is pointed to the virtual HBA driver's tasklet
   "(*funct)" function. The input argument is the address of mppLnx_taskletQueue_t 
   object.

* @taskletQueue the address of the mppLnxTaskletQueue_t object which has a list of all
*     deffered commands that must be processed.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_tasklet_func(unsigned long taskletQueue)
{
   mppLnx_taskletQueue_t         * taskQueuep = (mppLnx_taskletQueue_t *) taskletQueue;
   mppLnx_CmndEntry_t            * cmndEntry;
   struct list_head              * listPt;
   unsigned long                 iflages;
   MPP_TIME                      currentTime;
   MPP_TIME                      ControllerIoWaitTime;
   MPP_TIME                      ArrayIoWaitTime;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Entering mppLnx_tasklet_func()\n"));

   /*
   * lock the queue and process all of the commands in the queue
   */
   atomic_set(&taskQueuep->needSchedule,0);
   while(1)
   {
      spin_lock_irqsave(mppLnxTaskletQueue.cmndQueueLock, iflages);
      if( (listPt = taskQueuep->cmndEntry->list.next)
           != &(taskQueuep->cmndEntry->list) )
      {
         cmndEntry = list_entry(listPt, mppLnx_CmndEntry_t, list);
         spin_unlock_irqrestore(mppLnxTaskletQueue.cmndQueueLock, iflages);
         if(cmndEntry->taksType == MPPLNX_DELAY_RESPONSE_TYPE)
         {
            /*
            * completing delayed response command 
            */
            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "Completing scsi delayed virtual scsi_cmnd vtarget %d vlun %d\n",
               cmndEntry->vcmnd->target, cmndEntry->vcmnd->lun));
            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "the virtual command serial_number %ld command %x\n",
               cmndEntry->vcmnd->serial_number, cmndEntry->vcmnd->cmnd[0]));
            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "vcmnd->result %d\n", cmndEntry->vcmnd->result));
            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "mppLnx_tasklet_func cmnd %lx serial_number %ld\n",
               (unsigned long)cmndEntry->vcmnd, cmndEntry->vcmnd->serial_number));
            cmndEntry->vcmnd->scsi_done(cmndEntry->vcmnd);
         }
         else if(cmndEntry->taksType == MPPLNX_QUEUED_CMND_TYPE)
         {
            /*dispatch the proxy command to the physical HBA
            */
            MPP_GETTIME(currentTime);
            MPP_GETTIME(ControllerIoWaitTime);
            MPP_GETTIME(ArrayIoWaitTime);

            MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
               "queue the command to the physical HBA driver cmnd %02x %02x\n",
               cmndEntry->vcmnd->cmnd[0],
               cmndEntry->vcmnd->cmnd[1]));

            mppLnx_do_queuecommand(cmndEntry->vcmnd,
               middle_level_timeoutfunc,
               cmndEntry->vcmnd->scsi_done,
               UNKNOWN_PREVIOUS_CONTROLLER,
               UNKNOWN_PREVIOUS_PATH,
               0, /*CommandTimeoutRetryCount*/
               mppLnx_Config_Parameters.mpp_SyncRetryCount, /*RetryCount */
               0, /*SelectionTimeoutRetryCount*/
               0, /*UaRetryCount*/
               currentTime,   /*the current time */
               ControllerIoWaitTime, /*controller io wait start time*/
               ArrayIoWaitTime   /*the array io wait start time */
               );
         }
         else
         {
            /*
            * we should not get here
            */
            MPPLNX_VHBA_BUG();
         }
         /*remove the entry from the list*/
         spin_lock_irqsave(mppLnxTaskletQueue.cmndQueueLock, iflages);
         list_del(&cmndEntry->list);
         spin_unlock_irqrestore(mppLnxTaskletQueue.cmndQueueLock, iflages);
         kfree(cmndEntry);
      }else
      {
         spin_unlock_irqrestore(mppLnxTaskletQueue.cmndQueueLock, iflages);
         break;
      }
   }
   atomic_set(&taskQueuep->needSchedule,1);
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_tasklet_func()\n"));

}
#endif

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_Scsi2toScsi3_handler
* SUMMARY:  Handling Scsi-2 release/reserve virtual Scsi_Cmnd at a late time
* SCOPE:    public 
*
* DESCRIPTION:
   This function is called by mppLnx_queuecommand() function when a Scsi-2 reserve/release
   command is received by the mppLnx_queuecommand().
   The input argument is the Scsi-2 reserve/release command in the format of Scsi_Cmnd 
   dispatched by the Scsi middle level.
   This function simply creates a mppLnx_CmndEntry_t and inserts the object to
   mppLnxWorkerQueueContextContex. Then calls wakes up the sync IO worker thread to process
   the command

* @scp the virtual Scsi_Cmnd that is either Scsi-2 reserve or Scsi-2 release virtual command
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_Scsi2toScsi3_handler( Scsi_Cmnd * scp)
{
   mppLnx_CmndEntry_t      * scsi2CmndEntry;
   int                     size = 0;
   unsigned long           iflages;
   
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "Entering mppLnx_Scsi2toScsi3_handler()\n"));

   size = sizeof(mppLnx_CmndEntry_t);
   scsi2CmndEntry = kmalloc(size, GFP_ATOMIC);
   if( NULL == scsi2CmndEntry )
   {
      /* error. Cannot get memory */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,465,
         "Error could not allocate mppLnx_CmndEntry_t in mppLnx_Scsi2toScsi3_handler\n"));
      /*
      * complete the command with error
      */
      scp->result = DRIVER_ERROR << 24;
      scp->scsi_done(scp);
      return;
   }
   else
   {
      memset(scsi2CmndEntry, 0, size);
      spin_lock_irqsave(mppLnxWorkerQueueContextContext.cmndQueueLock, iflages);
      scsi2CmndEntry->vcmnd = scp;
      scsi2CmndEntry->taksType = MPPLNX_SCSI2TOSCSI3_TRAN_TYPE;
      list_add_tail(&(scsi2CmndEntry->list), &(mppLnxWorkerQueueContextContext.cmndEntry->list));
      spin_unlock_irqrestore(mppLnxWorkerQueueContextContext.cmndQueueLock, iflages);
   
      /*
       * delete the virtual command's timer so that the command will not time-out
       * We need to add the timer back before complete the virtual command
       */
      if(scp->eh_timeout.function != NULL)
      {
         scsi2CmndEntry->vcmnd_timeoutfunc = (void (*)(Scsi_Cmnd *))scp->eh_timeout.function;
         scsi_delete_timer(scp);
      }else
      {
         /*
         * this should be the case of the BDR message.
         */
         scsi2CmndEntry->vcmnd_timeoutfunc = NULL;
         MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
            "the vcmnd has no timer 0x%lx SNno 0x%lx cdb[0]= %x\n",
            (unsigned long)scp, scp->serial_number, scp->cmnd[0]));
      }

      if( atomic_read(&mppLnxWorkerQueueContextContext.worker_active) == 0)
      {
         MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_Scsi2toScsi3_handler()- wake up worker thread\n"));
         up(mppLnxWorkerQueueContextContext.workerQueue_sem);
      }
      else
      {
         MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
            "mppLnx_Scsi2toScsi3_handler()- worker thread is running\n"));
      }
   }
   MPP_DEBUGPRINT((MPP_RESERVE_DEBUG+MPP_DEBUG_LEVEL_1,
      "Exiting mppLnx_Scsi2toScsi3_handler()\n"));

}


/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_schedule_resp
* SUMMARY:  Schedule a Scsi Command to be completed at a late time
* SCOPE:   local
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This function is called by mppLnx_queuecommand() function when a non-dispatch-able 
   virtual command is received. The mppLnx_queuecommand() uses this function to defer 
   the non-dispatch-able command's completion.
   The input argument is the non-dispatch-able command in the format of Scsi_Cmnd.
   This function simply creates a mppLnx_CmndEntry_t and inserts the object to
   mppLnx_taskletQueue. It calls Linux kernel API tasklet_schedule() to schedule 
   the virtual HBA driver's tasklet.

* @ scp the input argument is the non-dispatch-able command in the format of Scsi_Cmnd.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_schedule_resp( Scsi_Cmnd * scp)
{
   mppLnx_CmndEntry_t      * delayedCmndEntry;
   int                     size = 0;
   unsigned long           iflages;

   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "mppLnx_schedule_resp received cmnd %lx serial_number %ld\n",
      (unsigned long)scp, scp->serial_number));

   size = sizeof(mppLnx_CmndEntry_t);
   delayedCmndEntry = kmalloc(size, GFP_ATOMIC);
   if( NULL == delayedCmndEntry )
   {
      /* error. Cannot get memory */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,466,
         "cannot get memory for mppLnx_CmndEntry_t\n"));
      /*
      * complete the command with error
      */
      scp->result = DRIVER_ERROR << 24;
      scp->scsi_done(scp);
      return;
   }
   else
   {
#if USE_THREAD_FOR_DPC
      spin_lock_irqsave(mppLnxDpcQueueContext.cmndQueueLock, iflages);
      memset(delayedCmndEntry, 0, size);
      delayedCmndEntry->vcmnd = scp;
      delayedCmndEntry->taksType = MPPLNX_DELAY_RESPONSE_TYPE;
      list_add_tail(&(delayedCmndEntry->list), 
			&(mppLnxDpcQueueContext.cmndEntry->list));
      spin_unlock_irqrestore(mppLnxDpcQueueContext.cmndQueueLock, iflages);
      if(atomic_read(&mppLnxDpcQueueContext.needSchedule))
      {
         up(mppLnxDpcQueueContext.dpc_sem);
      }
#else
      spin_lock_irqsave(mppLnxTaskletQueue.cmndQueueLock, iflages);
      memset(delayedCmndEntry, 0, size);
      delayedCmndEntry->vcmnd = scp;
      delayedCmndEntry->taksType = MPPLNX_DELAY_RESPONSE_TYPE;
      list_add_tail(&(delayedCmndEntry->list), 
			&(mppLnxTaskletQueue.cmndEntry->list));
      spin_unlock_irqrestore(mppLnxTaskletQueue.cmndQueueLock, iflages);
      if(atomic_read(&mppLnxTaskletQueue.needSchedule))
      {
         tasklet_schedule(mppLnxTaskletQueue.mppTasklet);
      }
#endif
   }
   MPP_DEBUGPRINT((MPP_DEBUG_LOG_ONLY+MPP_DEBUG_LEVEL_4,
      "Exiting mppLnx_schedule_resp\n"));
}

/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_schedule_queuecmd
* SUMMARY:  Schedule a Scsi Command to be completed at a late time
* SCOPE:   local
    local = Internal procedure (static)
    public = Procedure is visible to all components throughout system
    private = Procedure is visible to a limited set of modules, within
        one component (default)
*
*
* DESCRIPTION:
   This function is called by mppLnx_queuecommand() function when a dispatch-able 
   virtual command is received. The mppLnx_queuecommand() uses this function to dispatch
   the virtual command to the physical HBA driver.
   The input argument is the virtual command in the format of Scsi_Cmnd.
   This function simply creates a mppLnx_CmndEntry_t and inserts the object to
   mppLnx_taskletQueue. It calls Linux kernel API tasklet_schedule() to schedule 
   the virtual HBA driver's tasklet.

* @ scp the input argument is the non-dispatch-able command in the format of Scsi_Cmnd.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_schedule_queuecmd( Scsi_Cmnd * scp)
{
   mppLnx_CmndEntry_t      * queuedCmndEntry;
   int                     size = 0;
   unsigned long           iflages;

   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "mppLnx_schedule_queuecmd received cmnd %lx serial_number %ld\n",
      (unsigned long)scp, scp->serial_number));
  
   size = sizeof(mppLnx_CmndEntry_t);
   queuedCmndEntry = kmalloc(size, GFP_ATOMIC);
   if( NULL == queuedCmndEntry )
   {
      /* error. Cannot get memory */
      MPP_ERRORLOGPRINT((MPP_ERR_FATAL,467,
         "cannot get memory for mppLnx_CmndEntry_t\n"));
      /*
      * complete the command with error
      */
      scp->result = DRIVER_ERROR << 24;
      scp->scsi_done(scp);
      return;
   }
   else
   {
#if USE_THREAD_FOR_DPC
      spin_lock_irqsave(mppLnxDpcQueueContext.cmndQueueLock, iflages);
      memset(queuedCmndEntry, 0, size);
      queuedCmndEntry->vcmnd = scp;
      queuedCmndEntry->taksType = MPPLNX_QUEUED_CMND_TYPE;
      list_add_tail(&(queuedCmndEntry->list), 
			&(mppLnxDpcQueueContext.cmndEntry->list));
      spin_unlock_irqrestore(mppLnxDpcQueueContext.cmndQueueLock, iflages);

      if(atomic_read(&mppLnxDpcQueueContext.needSchedule))
      {
         up(mppLnxDpcQueueContext.dpc_sem);
      }
#else
      spin_lock_irqsave(mppLnxTaskletQueue.cmndQueueLock, iflages);
      memset(queuedCmndEntry, 0, size);
      queuedCmndEntry->vcmnd = scp;
      queuedCmndEntry->taksType = MPPLNX_QUEUED_CMND_TYPE;
      list_add_tail(&(queuedCmndEntry->list), 
			&(mppLnxTaskletQueue.cmndEntry->list));
      spin_unlock_irqrestore(mppLnxTaskletQueue.cmndQueueLock, iflages);

      if(atomic_read(&mppLnxTaskletQueue.needSchedule))
      {
         tasklet_schedule(mppLnxTaskletQueue.mppTasklet);
      }
#endif
   }
   
   MPP_DEBUGPRINT((MPP_HBA_COMMAND_DEBUG+MPP_DEBUG_LEVEL_2,
      "Exiting mppLnx_schedule_queuecmd\n"));
}
/*******************************************************************************
* PROCEDURE
*
* NAME:     mppLnx_setCheckCondition
* SUMMARY:  Prepare a SCSI command to return check-condition status
* SCOPE:   local
*
*
* DESCRIPTION:
   This function will set the command's result to CHECK_CONDITION and command's sense buffer
   to return 02/04/09 (Sense-key/ASC/ASCQ)

* @ vcmd the input argument. The virtual command needs to be completed
* @ sensedata - a SenseData_t object.
* RESTRICTIONS:
* 
* RETURNS:
*
* ERRNO:
*
* NOTES:
*
* EXAMPLES:
*
* SEE ALSO:
*
*/
void mppLnx_setCheckCondition(Scsi_Cmnd * vcmd, SenseData_t *sensedata)
{
   int senseLen = 0;
   
   senseLen = (sizeof(SenseData_t) < SCSI_SENSE_BUFFERSIZE)?
      sizeof(SenseData_t):SCSI_SENSE_BUFFERSIZE;
   /*
   *
   * only the first 18 bytes ( the SCSI standard sense data will be filled
   typedef struct SenseData {
	0 BYTE		Error_Code;
	1 BYTE		Segment_Number; // not applicable to DAC 
	2 BYTE		Sense_Key;
	3 BYTE		Information[ 4 ];
	7 BYTE	 	Additional_Len;
	8 LWORD		Command_Specific_Info;
	12 BYTE		ASC;
	13 BYTE		ASCQ;
   14 BYTE		Field_Replaceable_Unit;
	15 BYTE		Sense_Key_Specific_Info[ 3 ];
   */
   bzero(sensedata,sizeof(SenseData_t));
   sensedata->Error_Code = 0x70;
   sensedata->Segment_Number = 0x00;
   sensedata->Sense_Key = 0x02;
   /*
   * Information -> 4 bytes, all 0s
   */
   sensedata->Additional_Len = 0x98; /* 152 */
   /*
   *Command_Specific_Info. all 0s
   */
   sensedata->ASC = 0x04;
   sensedata->ASCQ = 0x09;
   /*
   *Field_Replaceable_Unit, 0. Sense_Key_Specific_Info, all 0s
   */
   
   /*
   * Scsi_Cmnd->results's status byte -> check-condition
   */
   vcmd->result = (CHECK_CONDITION<<1);

   bcopy (sensedata,vcmd->sense_buffer, senseLen );
}
