/*******************************************************************************
* Copyright (C) 2003 by LSI Logic Corporation.  
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation,  version 2 of the License.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
********************************************************************************/
/*******************************************************************************

NAME            MPPutil.c
SUMMARY         %description%
VERSION         %version: 27 %
UPDATE DATE     %date_modified: Mon Jan 19 14:53:24 2004 %
PROGRAMMER      %created_by:    shankara %

        Copyright 2000 by LSI Logic Inc.

DESCRIPTION:
	Utility program for the MPP driver.

USAGE:
	mppUtil [-d debug_level] [-e error_level] [-g target_id] [-s "failback" | "avt" | "preferred" | "busscan" ]
	
	where:
		-d 	Set debug level based on defines found in MPP_Common.h
		-e 	Set error reporting level based on defines found in
			MPP_Common.h
		-g 	Get the RdacInfo data from the specified target ID.
		-s 	Initiate a scan for failed controllers, AVT state
			change, or preferred path change.

*******************************************************************************/


/***  INCLUDES  ***/  
#include	"mppUtil_Sysdep.h"
#include	"MPP_Sysdep.h"
#include	"MPP_Common.h"
#include	"MPP_RdacInfo.h"
#include	"mppUtil_Common.h"

/***  EXTERNAL REFERENCES  ***/

/***  PROCEDURES  ***/
	int	__cdecl main(LINT argc, char **argv);
static	BOOL	utilSetDebugLevel(char *level);
static	BOOL	utilSetErrorLevel(LINT level);
static	BOOL	utilGetRequest(LINT target);
static	BOOL	utilScanRequest(char *req);
static	VOID	utilFormatTime(MPP_TIME rawTime, MPP_STRING buffer);
static	VOID	utilPrintRdacInfo(getRdacInfoIoctl_t *getInfo);
static	VOID	utilHexDump(unsigned char *buffer, int bufferLength);
static	BOOL	utilGetMaxLunsPerArray(void);
static	BOOL	utilGetMaxArrayModules(void);

/***  GLOBALS  ***/
	HANDLE   utilVirtualFd;
	LWORD    utilMaxPaths;
	LWORD    utilMaxLuns;
	BOOL     utilDebugEnabled;
	MPP_STRING  labelString;
	LINT    labelLength;
	CHAR	device[128];
	MPP_STRING progname;

int
__cdecl
main(LINT argc, char **argv)
{
	LINT	opt;
	BOOL	status = FALSE;
	char cmdLineArgs[MAX_CMDLINE_LENGTH];

	utilVirtualFd = (HANDLE) INVALID;
	utilMaxPaths = INVALID;
	utilMaxLuns = INVALID;
	utilDebugEnabled = FALSE;
	progname = argv[0];
	sprintf(cmdLineArgs, "%s%s", MPPCMN_UTIL_CMDLINE_ARGS,  MPPSYS_UTIL_CMDLINE_ARGS);
	while((opt = getopt(argc, argv, cmdLineArgs)) != EOF)
	{
		switch(opt)
		{
			case	'd':
				status = utilSetDebugLevel(optarg);
				break;

			case	'e':
				status = utilSetErrorLevel(strtol(optarg, (char **)NULL, 0));
				break;

			case	'g':
				status = utilGetRequest(atoi(optarg));
				break;

			case	's':
				status = utilScanRequest(optarg);
				break;

			case	't':
				utilDebugEnabled = TRUE;
				break;

			case	'o':
				strcpy(device, optarg);
				break;

			case	'm':
				status = utilGetMaxArrayModules();
				break;

			case	'l':
				status = utilGetMaxLunsPerArray();
				break;
				
			default:
				status = mppSys_ProcessCmdLineArgs(opt, optarg);
				break;
		}
		if ( status == INVALID )
			break;
	}

	return(status);
}


/*******************************************************************************
* PROCEDURE
*
* NAME:		utilUsage
* SUMMARY:	Display program usage.
* SCOPE:	LOCAL
*
* DESCRIPTION:
*	Display program usage.
*
* RETURNS:
*	N/A
*******************************************************************************/
VOID
utilUsage(char *prog)
{
	printf("Usage: %s %s %s\n", prog, MPPCMN_UTIL_USAGE, MPPSYS_UTIL_USAGE );
	return;
}


/*******************************************************************************
* PROCEDURE
*
* NAME:		utilSetDebugLevel
* SUMMARY:	Set the current MPP debug level.
* SCOPE:	LOCAL
*
* DESCRIPTION:
*	This function sends an MPP_IOCTL_SETDEBUG request to the MPP
*	virtual bus.
*
* RETURNS:
*	Boolean indicating success or failure of the request.
*******************************************************************************/
static BOOL
utilSetDebugLevel(char *level)
{
	unsigned int	debugLevel = 0, idx = 0;

	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	// See if the user began the debug level with '0x'
	if(!strncmp(level, "0x", 2))
	{
		level += 2;
	}

	while(level && *level && isxdigit(*level))
	{
		idx = *level++ - '0';
		if(9 < idx)
			idx -= 7;
		debugLevel <<= 4;
		debugLevel |= (idx & 0xff);
	}
	
	utilDebug("Set debug level to 0x%x\n", debugLevel);
	if(ioctl(utilVirtualFd, MPP_IOCTL_SETDEBUG, &debugLevel) == INVALID)
	{
		fprintf(stderr, "Unable to send debug request - %s!\n", strerror(errno));
		return(FALSE);
	}
	return(TRUE);
}


/*******************************************************************************
* PROCEDURE
*
* NAME:		utilSetErrorLevel
* SUMMARY:	Set the current MPP error level.
* SCOPE:	LOCAL
*
* DESCRIPTION:
*	This function sends an MPP_IOCTL_SETERROR request to the MPP
*	virtual bus.
*
* RETURNS:
*	Boolean indicating success or failure of the request.
*******************************************************************************/
static BOOL
utilSetErrorLevel(LINT level)
{
	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	utilDebug("Set error level to 0x%x\n", level);
	if(ioctl(utilVirtualFd, MPP_IOCTL_SETERROR, &level) == INVALID)
	{
		fprintf(stderr, "Unable to send error request - %s!\n", strerror(errno));
		return(FALSE);
	}
	return(TRUE);
}


/*******************************************************************************
* PROCEDURE
*
* NAME:		utilScanRequest
* SUMMARY:	Send a scan request to the MPP driver.
* SCOPE:	LOCAL
*
* DESCRIPTION:
*	This function sends a scan request to the MPP virtual bus.
*
* RETURNS:
*	Boolean indicating success or failure.
*******************************************************************************/
static BOOL
utilScanRequest(char *req)
{
	LINT	ioctlReq;

	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	if(strcmp(req, SCAN_FAILBACK) != 0 && strcmp(req, SCAN_AVT) != 0 &&
		strcmp(req, SCAN_PREFERRED) != 0 && strcmp(req, SCAN_BUSSCAN) != 0)
	{
		fprintf(stderr, "Invalid option!\n");
		return(FALSE);
	}

	if(!strcmp(req, SCAN_FAILBACK))
		ioctlReq = MPP_IOCTL_FAILBACKSCAN;
	else if(!strcmp(req, SCAN_AVT))
		ioctlReq = MPP_IOCTL_AVTCHECKSCAN;
	else if(!strcmp(req, SCAN_PREFERRED))
		ioctlReq = MPP_IOCTL_AVTPATHSCAN;
	else if(!strcmp(req, SCAN_BUSSCAN))
		ioctlReq = MPP_IOCTL_BUSSCAN;

	utilDebug("Send scan request 0x%x\n", ioctlReq);
	if(ioctl(utilVirtualFd, ioctlReq, NULL) == INVALID)
	{
		fprintf(stderr, "Unable to send ioctl request - %s!\n", strerror(errno));
		return(FALSE);
	}
	return(TRUE);
}


/*******************************************************************************
* PROCEDURE
*
* NAME:		utilGetRequest
* SUMMARY:	Get the RDAC information for a given target.
* SCOPE:	LOCAL
*
* DESCRIPTION:
*	This function sends a get request to the MPP virtual bus.
*
* RETURNS:
*	Boolean indicating success or failure.
*******************************************************************************/
static BOOL
utilGetRequest(LINT target)
{
	RdacDeviceInformation_t *reqInfo = NULL;
	unsigned char		reqName[64];
	getRdacInfoIoctl_t	rdacInfo;
	LWORD			Lun;
	BYTE			*reqPtr = NULL;

	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	if((reqInfo = (RdacDeviceInformation_t *)malloc(MPP_RDACINFOSIZE(utilMaxLuns, utilMaxPaths))) == NULL)
	{
		fprintf(stderr, "Unable to malloc memory for RdacInfo - %s!\n", strerror(errno));
		return(FALSE);
	}

	rdacInfo.target = target;
	rdacInfo.BufferSize = MPP_RDACINFOSIZE(utilMaxLuns, utilMaxPaths);
	rdacInfo.RdacInfoAdrs = reqInfo;
	rdacInfo.RdacName = &reqName;

	utilDebug("Send RdacInfo request\n");
	if(ioctl(utilVirtualFd, MPP_IOCTL_GETRDACINFO, &rdacInfo) == INVALID)
	{
		fprintf(stderr, "Unable to send ioctl request - %s!\n", strerror(errno));
		return(FALSE);
	}

	reqPtr = (BYTE *)reqInfo;
	reqPtr = reqPtr + sizeof(RdacDeviceInformation_t) +
		(sizeof(RdacLunInformation_t) * (utilMaxLuns - 1));
	reqInfo->ControllerInfo[0] = (RdacControllerInformation_t *)reqPtr;

	reqPtr = (reqPtr + sizeof(RdacControllerInformation_t) +
		(sizeof(RdacControllerPath_t) * (utilMaxPaths - 1)));
	reqInfo->ControllerInfo[1] = (RdacControllerInformation_t *)reqPtr;

	reqPtr = reqPtr + sizeof(RdacControllerInformation_t) +
		(sizeof(RdacControllerPath_t) * (utilMaxPaths - 1));
	for(Lun = 0; Lun < utilMaxLuns; ++Lun)
	{
		reqInfo->RdacLuns[Lun].LunControllers[0] = (LunPathObjects_t *)reqPtr;

		reqPtr = reqPtr + sizeof(LunPathObjects_t) +
			(sizeof(LunPathInfo_t) * (utilMaxPaths - 1));
		reqInfo->RdacLuns[Lun].LunControllers[1] = (LunPathObjects_t *)reqPtr;

		reqPtr = reqPtr + sizeof(LunPathObjects_t) +
			(sizeof(LunPathInfo_t) * (utilMaxPaths - 1));
	}

	labelString = (MPP_STRING) 0;
	labelLength = 0;
	utilPrintRdacInfo(&rdacInfo);
	return(FALSE);
}

static VOID
utilPrintRdacInfo(getRdacInfoIoctl_t *getInfo)
{
	int				controllerCount;
	LWORD				lunCount, pathCount;
	unsigned char			nullWWN[WWN_LENGTH];
	RdacDeviceInformation_t 	*rdacInfo;
	RdacControllerInformation_t	*controllerInfo;
	RdacControllerPath_t		*controllerPath;
	RdacLunInformation_t		*lunInfo;
	LunPathObjects_t		*lunPath;
	LunPathInfo_t			*lunPathInfo;
	char				strBuf[STR_LENGTH], strBuf2[STR_LENGTH];
	char				timeBuf[TIME_BUFFER_LENGTH];
	char				ctlAddressStrBuf[50] = "";

	rdacInfo = getInfo->RdacInfoAdrs;

	memset(nullWWN, '\0', WWN_LENGTH);

	/**
	 ** Global array information
	 **/
	printf("MPP Information:\n");
	printf("----------------\n");

	printf("      ModuleName: %-32.32s         SingleController: %c\n",
		getInfo->RdacName, YES_NO(rdacInfo->SingleController));
	printf(" VirtualTargetID: 0x%03x                                       ScanTriggered: %c\n",
		rdacInfo->VirtualTargetId, YES_NO(rdacInfo->ScanTriggered));
	printf("     ObjectCount: 0x%03x                                          AVTEnabled: %c\n",
		rdacInfo->ObjectCount, YES_NO(rdacInfo->AVTEnabled));
	printf("             WWN: ");
	utilHexDump(rdacInfo->WWN, WWN_LENGTH);
	printf("               RestoreCfg: %c\n", YES_NO(rdacInfo->RestoreCfg));
	utilFormatHandle(rdacInfo->ModuleHandle, strBuf);
	printf("    ModuleHandle: %-18.18s                          Page2CSubPage: %c\n",
		strBuf, YES_NO(rdacInfo->Page2CSubPage1));
	memset(strBuf, '\0', STR_LENGTH);
	sprintf(strBuf, "%x.%x.%x.%x", rdacInfo->FirmwareVersion[0], rdacInfo->FirmwareVersion[1], rdacInfo->FirmwareVersion[2], rdacInfo->FirmwareVersion[3]); 
	printf(" FirmwareVersion: %-18.18s                          \n",strBuf);
	
    printf("\n");

	/**
	 ** Controller-specific information
	 **/
	for(controllerCount = 0; controllerCount < 2; controllerCount++)
	{
		controllerInfo = rdacInfo->ControllerInfo[controllerCount];
		if(!controllerInfo->ControllerPresent)
		{
			continue;
		}

		printf("\nController '%c' Status:\n", controllerCount + 'A');
		printf("-----------------------\n");
		utilFormatHandle(controllerInfo->ControllerSlotHandle, strBuf);
		printf("ControllerHandle: %-18.18s                      ControllerPresent: %c\n",
			strBuf, YES_NO(controllerInfo->ControllerPresent));
		if(controllerInfo->UTMLunExists)
		{
			printf("    UTMLunExists: %c (%-3.3d)                                            Failed: %c\n",
				YES_NO(controllerInfo->UTMLunExists), controllerInfo->UTMLunNumber,
				YES_NO(controllerInfo->Failed));
		}
		else
		{
			printf("    UTMLunExists: %c                                                  Failed: %c\n",
				YES_NO(controllerInfo->UTMLunExists), YES_NO(controllerInfo->Failed));
		}
		printf("   NumberOfPaths: %1d                                          FailoverInProg: %c\n",
			controllerInfo->NumberOfPaths, YES_NO(controllerInfo->FailoverInProgress));

		for(pathCount = 0; pathCount < controllerInfo->NumberOfPaths; pathCount++)
		{
			controllerPath = &controllerInfo->ControllerPath[pathCount];
			printf("\n    Path #%1d\n", pathCount + 1);
			printf("    ---------\n");
			utilFormatVertex(controllerPath->DirectoryVertex, strBuf);
			printf(" DirectoryVertex: %-18.18s                                Present: %c\n", 
				strBuf, YES_NO(controllerPath->Present));
			printf("                                                                     Failed: %c\n",
				YES_NO(controllerPath->Failed));
			utilFormatControllerAddress(controllerPath->Address, ctlAddressStrBuf);
			printf(" %-42.42s", ctlAddressStrBuf );
		}

		printf("\n");
	}

	/**
	 ** Lun-specific information
	 **/

	printf("\n\nLun Information\n");
	printf("---------------\n");
	for(lunCount = 0; lunCount < utilMaxLuns; lunCount++)
	{
		lunInfo = &rdacInfo->RdacLuns[lunCount];

		if(!memcmp(nullWWN, lunInfo->WWN, WWN_LENGTH))
		{
			continue;
		}

		if(lunCount != 0)
		{
			printf("\n");
		}

		if(lunInfo->WWN[0] != 0)
		{
			printf("    Lun #%d - WWN: ", lunCount);
			utilHexDump(lunInfo->WWN, WWN_LENGTH);
			printf("\n");
		}
		else
		{
			printf("    Lun #%d - WWN:\n", lunCount);
		}

		printf("    ----------------\n");
		utilFormatVertex(lunInfo->PseudoLunObject, strBuf),
		printf("       LunObject: %-18.18s                      CurrentOwningPath: %c\n", strBuf,
			A_OR_B(lunInfo->CurrentOwningPath));
		printf("  RemoveEligible: %c                                          BootOwningPath: %c\n",
			YES_NO(lunInfo->PseudoRemoveEligible), A_OR_B(lunInfo->BootOwningPath));
		printf("   NotConfigured: %c                                           PreferredPath: %c\n",
			YES_NO(lunInfo->NotConfigured), A_OR_B(lunInfo->PreferredPath));
		printf("        DevState: %-21.21s                     ReportedPresent: %c\n",
			utilFormatDevState(lunInfo->PseudoDevState), YES_NO(lunInfo->ReportedPresent));
		printf("    PrevDevState: %-21.21s                     ReportedMissing: %c\n",
			utilFormatDevState(lunInfo->PseudoPreviousDevState), YES_NO(lunInfo->ReportedMissing));
		if(lunInfo->NotReadyStart)
		{
			utilFormatTime(lunInfo->NotReadyStart, timeBuf);
			printf("  NotReady Start: %-19.19s                               NotReady: %c\n",
				timeBuf, YES_NO(lunInfo->NotReady));
		}
		else
		{
			printf("                                                                   NotReady: %c\n",
				YES_NO(lunInfo->NotReady));
		}

		if(lunInfo->BusyStart)
		{
			utilFormatTime(lunInfo->BusyStart, timeBuf);
			printf("      Busy Start: %-19.19s                                     Busy: %c\n",
				timeBuf, YES_NO(lunInfo->Busy));
		}
		else
		{
			printf("                                                                       Busy: %c\n",
				YES_NO(lunInfo->Busy));
		}

		if(lunInfo->QuiescienceStart)
		{
			utilFormatTime(lunInfo->QuiescienceStart, timeBuf);
			printf("Quiescience Start: %-19.19s                            Quiescent: %c\n\n",
				timeBuf, YES_NO(lunInfo->Quiescent));
		}
		else
		{
			printf("                                                                  Quiescent: %c\n\n", YES_NO(lunInfo->Quiescent));
		}

		for(controllerCount = 0; controllerCount < 2; controllerCount++)
		{
			if(!lunInfo->LunControllers[controllerCount])
			{
				continue;
			}

			lunPath = lunInfo->LunControllers[controllerCount];
			printf("    Controller '%c' Path\n", controllerCount + 'A');
			printf("    --------------------\n");
			printf("   NumLunObjects: %1d                                         RoundRobinIndex: %1d\n",
				lunPath->NumberOfLunObjects, lunPath->RoundRobinPathIndex);

			//for(pathCount = 0; pathCount < lunPath->NumberOfLunObjects; pathCount++)
			// 08/07/03 - Changed NumberOfLunObjects to utilMaxPaths.  NumberOfLunObjects reflects the current number
			// of lun paths seen for a given lun.  Under Windows PnP a given lun may not have the same number of lun
			// paths as another lun.  For example, Lun 0 may have 4 paths to each controller but Lun 1 may only have
			// 3 paths (due to some problem with that lun on some path).  As a result, Lun 3 may have paths 1,2, and 4
			// defined.  If NumberOfLunObjects is 3 it will only print the first 3 slots of the LunPaths[] structure
			// (ie. 1,2, and 3).  Since path 4 doesn't print the user is not aware another path exists.  Changing the
			// logic to use MaxPathsPerController, and checking for UtmLunDevice or LunPathDevice is a better alternative.
			for(pathCount = 0; pathCount < utilMaxPaths; pathCount++)
			{
				lunPathInfo = &lunPath->LunPaths[pathCount];
				if(!(lunPathInfo->UtmLunDevice || lunPathInfo->LunPathDevice))
				{
					continue;
				}

				utilFormatVertex(lunPathInfo->LunPathDevice, strBuf);

			printf("     RemoveState: 0x%x StartState: 0x%x PowerState: 0x%x\n", lunPathInfo->PathRemoveState, lunPathInfo->PathStartState, lunPathInfo->PathPowerState);
                if(lunPathInfo->UtmLunDevice != 0)
				{
					utilFormatVertex(lunPathInfo->UtmLunDevice, strBuf2),
					printf("         Path #%1d: LunPathDevice: %-18.18s            UTMLunDevice: %-18.18s\n",
                           pathCount + 1, strBuf, strBuf2);
				}
				else
				{
					printf("         Path #%1d: LunPathDevice: %-18.18s\n",
						pathCount + 1, strBuf);
				}

				printf("                        IoCount: %d\n", lunPathInfo->IoCount);
			}
			printf("\n");
		}
	}
}

static BOOL
utilGetMaxArrayModules( ) 
{
	LWORD maxArrayModules = 0;
	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	utilDebug("sending ioctl for maxArrayModules \n");
	if((ioctl(utilVirtualFd, MPP_IOCTL_GETMODULES, &maxArrayModules)) == INVALID)
	{
		fprintf(stderr, "Unable to get maxmodules request - %s!\n", strerror(errno));
		return(FALSE);
	}
	printf("MaxArrayModules = %d\n", maxArrayModules);
	return(TRUE);

}

static BOOL
utilGetMaxLunsPerArray( ) 
{
	LWORD maxLunsPerArray = 0;
	if(!utilDetermineBusFd())
	{
		fprintf(stderr, "Unable to determine virtual bus!!\n");
		return(INVALID);
	}

	utilDebug("sending ioctl for maxLunsPerArray\n");
	if((ioctl(utilVirtualFd, MPP_IOCTL_GETLUNS, &maxLunsPerArray)) == INVALID)
	{
		fprintf(stderr, "Unable to get maxmodules request - %s!\n", strerror(errno));
		return(FALSE);
	}
	printf("MaxLunsPerArray = %d\n", maxLunsPerArray);
	return(TRUE);

}



static VOID
utilFormatTime(MPP_TIME rawTime, MPP_STRING buffer)
{
	tm_t	*tmPtr;
	//struct tm	*tmPtr;
	//struct _SYSTEMTIME *tmPtr;

	memset(buffer, '\0', TIME_BUFFER_LENGTH);

	tmPtr = gmtime((time_t *) &rawTime);

	if(tmPtr)
	{
		sprintf(buffer, "%02d/%02d/%04d %02d:%02d:%02d",
			tmPtr->tm_mon, tmPtr->tm_mday, tmPtr->tm_year + 1900, 
			tmPtr->tm_hour, tmPtr->tm_min, tmPtr->tm_sec);
	}
	else
		sprintf(buffer, "Format error");

	return;
}

static VOID
utilHexDump(unsigned char *buffer, int bufferLength)
{
	int		cnt;

	for(cnt = 0; cnt < bufferLength; cnt++)
		printf("%02x", buffer[cnt]);
}


TINY	debugBuffer[DEBUG_BUFFER_LENGTH];

/*******************************************************************************
* PROCEDURE
*
* NAME:		utilDebug
* SUMMARY:	Display debug message.
* SCOPE:	GLOBAL
*
* DESCRIPTION:
*	Display debug message.
*
* RETURNS:
*	N/A
*******************************************************************************/
VOID
utilDebug(TINY *debugMessage, ...)
{
	va_list	ap;

	if(utilDebugEnabled)
	{
		va_start(ap, debugMessage);
		vsprintf(debugBuffer, debugMessage, ap);
		va_end(ap);
		fprintf(stderr, debugBuffer);
	}
	return;
}
